# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from collections.abc import Callable

import numpy as np
from numpy import ndarray
from mne.utils import verbose, logger

from .base import Transforms, TransformsEGD
from ..datasets.base import EEGData
from ..utils import DPEEG_SEED, get_init_args
from .functions import (
    crop,
    slide_win,
    cheby2_filter,
    label_mapping,
    pick_label,
)


class Identity(Transforms):
    """Placeholder identity operator."""

    def __init__(self) -> None:
        super().__init__("Identity()")

    def _apply(self, eegdata):
        return eegdata


class Crop(TransformsEGD):
    """Crop a time interval.

    Crop the eeg signal in terms of time. Default is `edata`.

    Parameters
    ----------
    tmin : int
        Start time of selection in sampling points.
    tmax : int
        End time of selection in sampling points. None means use the full time.
    include_tmax : bool
        If `False`, exclude tmax.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.Crop(4, 9)(eegdata, verbose=False)
    [edata=(16, 3, 5), label=(16,)]
    """

    def __init__(
        self,
        tmin: int = 0,
        tmax: int | None = None,
        include_tmax: bool = False,
    ) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.tmin = tmin
        self.tmax = tmax
        self.include_tmax = include_tmax

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd["edata"] = crop(
            data=egd["edata"],
            tmin=self.tmin,
            tmax=self.tmax,
            include_tmax=self.include_tmax,
        )


class SlideWin(TransformsEGD):
    """Apply a sliding window to the dataset.

    This transform is only splits the time series (dim = -1) through the
    sliding window operation on the original dataset. If the time axis is
    not divisible by the sliding window, the last remaining time data will
    be discarded. Applied to `edata` and `label` by default.

    Parameters
    ----------
    win : int
        The size of the sliding window (sampling point).
    overlap : int
        The amount of overlap between adjacent windows (sampling point).

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.SlideWin(3, 1)(eegdata, verbose=False)
    [edata=(64, 3, 3), label=(64,)]
    """

    def __init__(self, win: int, overlap: int = 0) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.win = win
        self.overlap = overlap

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd["edata"], egd["label"] = slide_win(
            data=egd["edata"],
            win=self.win,
            overlap=self.overlap,
            label=egd["label"],
        )


class Unsqueeze(TransformsEGD):
    """Insert a dimension on the data.

    This transform is usually used to insert a empty dimension on signals.

    Parameters
    ----------
    key : str
        The key of the eegdata value to be transformed.
    dim : int
        Position in the expanded dim where the new dim is placed.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.Unsqueeze(dim=2)(eegdata, verbose=False)
    [edata=(16, 3, 1, 10), label=(16,)]
    """

    def __init__(self, key: str = "edata", dim: int = 1) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.key = key
        self.dim = dim

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd[self.key] = np.expand_dims(egd[self.key], self.dim)


class Squeeze(TransformsEGD):
    """Remove a dimension on the data.

    Parameters
    ----------
    key : str
        The key of the eegdata value to be transformed.
    dim : int
        Selects a subset of the entries of length one in the shape. If a dim is
        selected with shape entry greater than one, an error is raised.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 1, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.Squeeze()(eegdata, verbose=False)
    [edata=(16, 3, 10), label=(16,)]
    """

    def __init__(self, key: str = "edata", dim: int = 1) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.key = key
        self.dim = dim

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd[self.key] = np.squeeze(egd[self.key], self.dim)


class Transpose(TransformsEGD):
    """Data dims transposed.

    By default, the EEG data (``edata``) of eegdata are transposed.

    Parameters
    ----------
    dims : tuple or list of int, optinal
        A tuple or list contains a permutation of [0,1,...,N-1] where N is the
        number of dims of the key values. The ``i`` ^th dim of the value will
        correspond to the axis numbered ``dims[i]`` of the input. If not
        specified, reverse the dims order by default.
    key : str
        The key of the eegdata value to be transformed.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 1, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.Transpose()(eegdata, verbose=False)
    [edata=(10, 3, 16), label=(16,)]
    """

    def __init__(self, dims: list[int] | None = None, key: str = "edata"):
        super().__init__(get_init_args(self, locals(), "rp"))
        self.dims = dims
        self.key = key

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd[self.key] = np.transpose(egd[self.key], self.dims)


class FilterBank(TransformsEGD):
    """Filter Bank.

    EEG data will be filtered according to different filtering frequencies and
    finally concatenated together. eg.`(Batch, ...) -> (Batch, F, ...)` if the
    number of filter banks exceeds 1, `(Batch, ...) -> (Batch, ...)` if the
    filter has only one. By default, filtering is performed on `edata`, please
    ensure the availability of the data. Related references include [1]_ and
    [2]_.

    Parameters
    ----------
    freq : float
        EEG data sampling frequency.
    filter_bank : multiple 2 float of list
        The low-pass and high-pass cutoff frequencies for each filter set.
    transition_bandwidth : float
        The bandwidth (in hertz) of the transition region of the frequency
        response from the passband to the stopband.
    gstop : float
        The minimum attenuation in the stopband (dB).
    gpass : float
        The maximum loss in the passband (dB).

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    References
    ----------
    .. [1] R. Mane, E. Chew, K. Chua, K. K. Ang, N. Robinson, A. P. Vinod,
        S.-W. Lee, and C. Guan, “FBCNet: A multi-view convolutional neural
        network for brain-computer interface,” arXiv preprint arXiv:2104.01233,
        2021.
    .. [2] X. Ma, W. Chen, Z. Pei, J. Liu, B. Huang, and J. Chen, “A temporal
        dependency learning CNN with attention mechanism for MI-EEG decoding,”
        IEEE Transactions on Neural Systems and Rehabilitation Engineering,
        2023.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.FilterBank(250)(eegdata, verbose=False)
    [edata=(16, 9, 3, 10), label=(16,)]
    """

    def __init__(
        self,
        freq: float,
        filter_bank: list = [
            [4, 8],
            [8, 12],
            [12, 16],
            [16, 20],
            [20, 24],
            [24, 28],
            [28, 32],
            [32, 36],
            [36, 40],
        ],
        transition_bandwidth: float = 2.0,
        gstop: float = 30,
        gpass: float = 3,
    ) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.freq = freq
        self.filter_bank = self._check_filter_bank(filter_bank)
        self.transition_bandwidth = transition_bandwidth
        self.gpass = gpass
        self.gstop = gstop
        self.bank_len = len(self.filter_bank)

    def _check_filter_bank(self, fb):
        if not isinstance(fb, list):
            raise TypeError(f"filter_bank must be a list, not {type(fb)}.")
        for f in fb:
            if len(f) != 2:
                raise ValueError(
                    "The filter should be of two variables low pass and high "
                    "pass cutoff frequency."
                )
        return fb

    def _apply_egd(self, egd: EEGData, key: str | None):
        trials = egd.trials()
        data = np.empty((trials, self.bank_len, *egd["edata"].shape[1:]))

        for i, cutoff in enumerate(self.filter_bank):
            filter_data = cheby2_filter(
                data=egd["edata"],
                freq=self.freq,
                l_freq=cutoff[0],
                h_freq=cutoff[1],
                transition_bandwidth=self.transition_bandwidth,
                gpass=self.gpass,
                gstop=self.gstop,
            )
            data[:, i] = filter_data

        if self.bank_len == 1:
            data = np.squeeze(data, 1)
        egd["edata"] = data


class ApplyFunc(TransformsEGD):
    """Apply a custom function to data.

    Parameters
    ----------
    func : Callable
        Transformation data callback function. The first parameter of the
        function must be `EEGData`.
    keys : list of str, optional
        The key of the eegdata to be transformed, if required. Applies to all
        eegdata by default.
    **kwargs : dict, optional
        Additional arguments for callback function, if required.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    If you want to pass a function with parameters, such as you want to use
    `np.expand_dims()` with `axis` parameter, you can do as follows:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> def expand_dim(data, dim=1):
    ...     data["edata"] = np.expand_dims(data["edata"], dim)
    >>> transforms.ApplyFunc(expand_dim, dim=0)(eegdata, verbose=False)
    [edata=(1, 16, 3, 10), label=(16,)]

    >>> split_eegdata = dpeeg.SplitEEGData(eegdata, eegdata.copy())
    >>> transforms.ApplyFunc(expand_dim, ["train"])(split_eegdata, verbose=False)
    Train: [edata=(1, 1, 16, 3, 10), label=(16,)]
    Test : [edata=(1, 16, 3, 10), label=(16,)]
    """

    def __init__(
        self,
        func: Callable,
        keys: list[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.func = func
        self.keys = keys
        self.kwargs = kwargs

    def _apply_egd(self, egd: EEGData, key: str | None):
        if (self.keys is None) or (key in self.keys):
            self.func(egd, **self.kwargs)


class LabelMapping(TransformsEGD):
    """Update the original label according to mapping rules.

    Parameters
    ----------
    mapping : ndarray (2, label_num), optional
        Label mapping relationship. The first row is the original label, and
        the second row is the mapped label. If ``None``, the label will be
        updated in ascending order starting from zero.
    order : bool
        Force the new labels to start incrementing from zero.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> eegdata['label']
    array([3, 2, 2, 2, 3, 2, 4, 3, 4, 3, 3, 2, 4, 4, 2, 3])

    Merge labels as needed:

    >>> transforms.LabelMapping(
    ...     np.array([[2, 3, 4], [0, 0, 1]])
    ... )(eegdata, verbose=False)
    >>> eegdata["label"]
    array([0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0])
    """

    def __init__(self, mapping: ndarray | None = None, order: bool = True) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.mapping = mapping
        self.order = order

    def _apply_egd(self, egd: EEGData, key: str | None):
        egd["label"] = label_mapping(
            label=egd["label"],
            mapping=self.mapping,
            order=self.order,
        )


class PickLabel(TransformsEGD):
    """Pick a subset of data.

    Pick the required labels and data from the dataset and re-label them.

    Parameters
    ----------
    pick : ndarray
        Label to include.
    keys : list of str, optional
        The key of the eegdata value to be transformed, if required. Applies to
        all eegdata by default.
    order : bool
        If `True`, relabel the selected labels.
    shuffle : bool
        Whether or not to shuffle the data after picking.
    seed : int
        Controls the shuffling applied to the data after picking.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    array([1, 2, 0, 2, 1, 2, 0, 1, 0, 0, 0, 1, 2, 1, 0, 0])

    >>> transforms.PickLabel(np.array([1, 2]))(eegdata, verbose=False)
    array([1, 0, 1, 0, 1, 0, 0, 0, 1])

    If some values do not need to be transformed, they can be excluded by the
    `keys` parameter:

    >>> eegdata = dpeeg.EEGData(
    ...     edata=np.random.randn(16, 3, 10),
    ...     label=np.random.randint(0, 3, 16),
    ...     adj=np.random.randn(16, 3, 3),
    ...     pcc=np.random.randn(16, 3, 3),
    ... )
    >>> transforms.PickLabel(
    ...    np.array([0, 1]), keys=["edata", "adj"]
    ... )(eegdata, verbose=False)
    [edata=(12, 3, 10), label=(12,), adj=(12, 3, 3), pcc=(16, 3, 3)]
    """

    def __init__(
        self,
        pick: ndarray,
        keys: list[str] | None = None,
        order: bool = True,
        shuffle: bool = True,
        seed: int = DPEEG_SEED,
    ) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        self.pick = pick
        self.keys = keys
        self.order = order
        self.shuffle = shuffle
        self.seed = seed

    def _apply_egd(self, egd: EEGData, key: str | None):
        label = egd["label"]

        keys, values = [], []
        for key, value in egd.items():
            if (key != "label") and ((self.keys is None) or (key in self.keys)):
                keys.append(key)
                values.append(value)

        data, label = pick_label(
            *values,
            label=label,
            pick=self.pick,
            order=self.order,
            shuffle=self.shuffle,
            seed=self.seed,
        )

        egd["label"] = label
        for i, key in enumerate(keys):
            egd[key] = data[i]



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from email.policy import strict
from pathlib import Path
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from typing import Any, TypeAlias, TypeVar, overload
from copy import deepcopy

import mne
import numpy as np
from mne.io import Raw
from mne import Epochs
from tqdm import tqdm
from numpy import ndarray

from ..utils import (
    get_init_args,
    iterable_to_str,
    mapping_to_str,
    _format_log,
    DPEEG_DIR,
)


class BaseData(ABC):
    """The bottom-level base data class."""

    @abstractmethod
    def _datas(self) -> Generator[tuple["EEGData", Any], None, None]:
        pass

    def copy(self):
        """Creates a deep copy of the current object."""
        return deepcopy(self)


_BaseData = TypeVar("_BaseData", bound=BaseData)


class EEGData(BaseData):
    """The base eegdata class.

    The most basic eeg data wrapper. It is essentially a python ``dict``, but
    with some additional functions added to it.

    Parameters
    ----------
    edata : array shape (n_trials, ...), optional
        EEG data.
    label : array shape (n_trials, ...), optional
        The labels corresponding to the eeg data.
    strict : bool
        ``True`` means that the number of ``edata`` and ``label`` samples must
        be the same.
    **kwargs : dict, optional
        Other parameter indicators.

    Examples
    --------
    If no value is passed, an empty EEGData object is initialized:

    >>> dpeeg.EEGData()
    [edata=(0,), label=(0,)]

    Or initialize with additional parameters:

    >>> dpeeg.EEGData(
    ...     edata=np.random.rand(16, 3, 20),
    ...     label=np.random.randint(0, 3, 16),
    ...     adj=np.random.rand(3, 3)
    ... )
    [edata=(16, 3, 20), label=(16,), adj=(3, 3)]
    """

    def __init__(
        self,
        edata: ndarray | None = None,
        label: ndarray | None = None,
        strict: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        edata = np.empty(0) if edata is None else edata
        label = np.empty(0) if label is None else label
        if strict:
            self._check(edata, label)

        self.data = {"edata": edata, "label": label, **kwargs}

    def _check(self, edata, label, raise_error=True):
        if len(edata) != len(label):
            if raise_error:
                raise ValueError(
                    f"The length {len(edata)} of edata is not equal to "
                    f"the length {len(label)} of label."
                )
            else:
                return False

        return True

    def __getitem__(self, key) -> ndarray:
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return self.data.keys()

    def values(self):
        """D.values() -> an object providing a view on D's values"""
        return self.data.values()

    def items(self):
        """D.items() -> a set-like object providing a view on D's items"""
        return self.data.items()

    def pop(self, key, default, /):
        """D.pop(k[,d]) -> v, remove specified key and return the value

        If the key is not found, return the default if given; otherwise,
        raise a KeyError.
        """
        if key in ["edata", "label"]:
            raise KeyError(f"{key} as a basic structure cannot be popped")
        self.data.pop(key, default)

    def __len__(self):
        """Returns the number of internal keys, not the number of samples."""
        return len(self.data)

    def check(self) -> bool:
        """Check whether the number of samples of the current `edata` and
        `label` is equal.
        """
        return self._check(self.data["edata"], self.data["label"], False)

    def trials(self) -> int:
        """Returns the number of trial samples."""
        return len(self.data["label"])

    def index(self, idx: Iterable[int] | slice) -> "EEGData":
        """Index internal data and return a new EEGData instance.

        Parameters
        ----------
        idx : Iterable[int], slice
            The index to apply to the internal data.

        Returns
        -------
        EEGData
            A new EEGData instance containing the indexed data.

        Examples
        --------
        >>> eegdata = dpeeg.EEGData(
        ...     edata=np.random.randn(16, 3, 10),
        ...     label=np.random.randint(0, 3, 16),
        ... )
        >>> eegdata.index([1, 2, 3])
        [edata=(3, 3, 10), label=(3,)]

        >>> eegdata.index(slice(6))
        [edata=(6, 3, 10), label=(6,)]
        """
        eegdata = EEGData()
        for key, value in self.data.items():
            eegdata[key] = value[idx]
        return eegdata

    def _datas(self) -> Generator[tuple["EEGData", None], None, None]:
        yield self, None

    def append(
        self, eegdata: "EEGData", dims: list[int] | int = 0, ignore: bool = False
    ):
        """Append a new eegdata at the end of the eegdata.

        Parameters
        ----------
        eegdata : EEGData
            The new eegdata to be added. If the new eegdata contains key-value
            pairs that do not exist in the original eegdata, they will be
            ignored by default.
        dims : int, list of int
            The dimension along which the new eegdata is merged with the old
            eegdata. If a single number is entered, all eegdata is added along
            that dimension. If a list is given, the new eegdata is added along
            the specified dimensions in the order given in the list and
            dictionary.
        ignore : bool
            If ``True``, add new key-value pairs from the new eegdata.

        Examples
        --------
        >>> eegdata = dpeeg.EEGData(
        ...     edata=np.random.randn(16, 3, 10),
        ...     label=np.random.randint(0, 3, 16),
        ... )
        >>> eegdata.append(eegdata.copy())
        >>> eegdata
        [edata=(32, 3, 10), label=(32,)]

        >>> eegdata_adj = eegdata.copy()
        >>> eegdata_adj["adj"] = np.random.randn(16, 3, 3)
        >>> eegdata_adj
        [edata=(32, 3, 10), label=(32,), adj=(16, 3, 3)]

        >>> eegdata.append(eegdata_adj, ignore=True)
        >>> eegdata
        [edata=(64, 3, 10), label=(64,), adj=(16, 3, 3)]
        """
        if isinstance(dims, int):
            dims = [dims] * len(eegdata)

        keys = self.keys()
        for i, key in enumerate(eegdata.keys()):
            if (key in keys) or (not ignore):
                self.add(key, eegdata[key], dim=dims[i])
            else:
                self[key] = eegdata[key]

    def add(self, key: str, value: ndarray | list[ndarray], dim: int = 0):
        """Add a new value to the corresponding key.

        Parameters
        ----------
        key : str
            The key of eegdata.
        value : ndarray, list of ndarray
            The new value to be added. If a `list`, all values in the list are
            added.
        dim : int
            The dimension in which the new data is concatenated with the old
            data.

        Examples
        --------
        >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
        ...                         label=np.random.randint(0, 3, 16))
        >>> eegdata.add("edata", np.random.randn(16, 3, 10), dim=2)
        >>> eegdata
        [edata=(16, 3, 20), label=(16,)]
        """
        if len(self[key]) == 0:
            if isinstance(value, list):
                self[key] = np.concatenate(value, axis=dim)
            else:
                self[key] = value

        else:
            data = [self[key]]
            if isinstance(value, ndarray):
                data.append(value)
            else:
                data.extend(value)
            self[key] = np.concatenate(data, axis=dim)

    @property
    def cls(self) -> list[int]:
        """Returns all categories."""
        return list(np.unique(self.data["label"]))

    @property
    def ncls(self) -> int:
        """Returns the number of categories."""
        return len(np.unique(self.data["label"]))

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the ``edata`` shape."""
        return self.data["edata"].shape[1:]

    def __repr__(self) -> str:
        data = {key: value.shape for key, value in self.items()}
        return f"[{mapping_to_str(data)}]"


class MultiSessEEGData(BaseData, dict):
    """Multi-session eegdata.

    The multi-session EEGData is actually equivalent to the EEGData collection
    wrapped by the Python built-in ``dict``. It does not have to be multi-
    session data, it can be any other EEGData collection. This name is only
    used here to indicate this type of data. The actual format is similar to:

        multi_sess_eegdata = { 'session_id': EEGData }

    .. Warning::

       Do not set the key value to the keywords "train" or "test"; otherwise,
       please use :class:`~SplitEEGData`.

    Parameters
    ----------
    data : list or dict of EEGData
        Initialize EEG data. If it is a list of ``EEGData``, add a key to each
        data in the order of the list. If it is a dict of ``EEGData``, use the
        dict to initialize.

    Examples
    --------
    Data type is list:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> dpeeg.MultiSessEEGData([eegdata.copy(), eegdata.copy()])
    {'session_1': [edata=(16, 3, 10), label=(16,)],
     'session_2': [edata=(16, 3, 10), label=(16,)]}

    or the data type is custom dict:

    >>> dpeeg.MultiSessEEGData({'xx': eegdata.copy(), 'yy': eegdata.copy()})
    {'xx': [edata=(16, 3, 10), label=(16,)],
     'yy': [edata=(16, 3, 10), label=(16,)]}
    """

    @overload
    def __init__(self, data: list[EEGData]) -> None:
        pass

    @overload
    def __init__(self, data: dict[str, EEGData]) -> None:
        pass

    def __init__(self, data):
        if isinstance(data, list):
            data = {f"session_{i + 1}": session for i, session in enumerate(data)}
        elif isinstance(data, dict):
            pass
        else:
            raise TypeError(f"Input type {type(data)} cannot be parsed.")

        super().__init__(data)

    def __getitem__(self, key: str) -> EEGData:
        return super().__getitem__(key)

    def trials(self, sum: bool = True) -> list[int] | int:
        """Returns the sum of the number of trials for all data or the number
        of trials for each data.
        """
        trials = []
        for session in self.keys():
            trials.append(len(self[session]["label"]))
        if sum:
            return np.sum(trials)
        else:
            return trials

    def _datas(self) -> Generator[tuple[EEGData, str], None, None]:
        for session, eegdata in self.items():
            yield eegdata, session


class SplitEEGData(BaseData):
    """Split eegdata.

    The eegdata has been split into training and test sets. A dict of data is
    maintained internally.

    Parameters
    ----------
    train : EEGData
        Training set data.
    test : EEGData
        Test set data.
    strict : bool
        The training and test sets `edata` are required to have the same shape.
    """

    def __init__(self, train: EEGData, test: EEGData, strict: bool = True) -> None:
        super().__init__()

        if strict and (train.shape != test.shape):
            raise ValueError(
                f"The training data shape {train.shape} is not the same as "
                f"the test data shape {test.shape}."
            )

        self.data = {"train": train, "test": test}

    def __setitem__(self, key: str, value: EEGData) -> None:
        if key not in ["train", "test"]:
            raise KeyError("The key must be train or test")
        self.data[key] = value

    def __getitem__(self, key: str) -> EEGData:
        return self.data[key]

    def trials(self) -> tuple[int, int]:
        """Returns the number of trials for training and testing data
        respectively.
        """
        train_trials = len(self.data["train"]["label"])
        test_trials = len(self.data["test"]["label"])
        return train_trials, test_trials

    def _datas(self) -> Generator[tuple[EEGData, str], None, None]:
        for mode, eegdata in self.data.items():
            yield eegdata, mode

    def __repr__(self) -> str:
        return f"Train: {self.data['train']}\n" f"Test : {self.data['test']}"


_DataAlias: TypeAlias = EEGData | MultiSessEEGData | SplitEEGData
_DataVar = TypeVar("_DataVar", EEGData, MultiSessEEGData, SplitEEGData)


class BaseDataset(ABC):
    """The bottom-level base dataset class."""

    def __init__(
        self,
        repr: dict,
        event_id: dict[str, int] | None = None,
    ) -> None:
        super().__init__()
        self.event_id = event_id
        self._repr = repr

    @abstractmethod
    def keys(self) -> list[int]:
        pass

    @abstractmethod
    def _get_single_subject_data(self, subject: int) -> _DataVar:
        pass

    def __getitem__(self, subject: int):
        """Get the subject's eegdata."""
        return self._get_single_subject_data(subject)

    def __len__(self) -> int:
        return len(self.keys())

    def values(self):
        for subject in self.keys():
            yield self._get_single_subject_data(subject)

    def items(self):
        for subject in self.keys():
            subject_data = self._get_single_subject_data(subject)
            yield subject, subject_data

    def __repr__(self) -> str:
        return _format_log(self._repr)


_BaseDataset = TypeVar("_BaseDataset", bound=BaseDataset)


class EEGDataset(BaseDataset):
    """Base EEG dataset.

    Parameters
    ----------
    eegdataset : list or dict of eegdata, optional
        The eegdata of different subjects. If ``None``, initialize an empty
        dataset. If ``list``, sort by subject one.
    event_id : dict, optional
        The correspondence between labels and events.
    rename : str, optional
        User renamed name. If ``None``, the default class name is used.

    Notes
    -----
    The dataset supports different subjects with different eegdata types (such
    as ``EEGData``, ``MultiSessEEGData`` and ``SplitEEGData``), but it is
    recommended to unify the eegdata types of all subjects when performing
    transformation and training models to avoid unpredictable errors.
    """

    @overload
    def __init__(
        self,
        eegdataset: list[BaseData] | None = None,
        event_id: dict[str, int] | None = None,
        rename: str | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        eegdataset: dict[int, BaseData] | None = None,
        event_id: dict[str, int] | None = None,
        rename: str | None = None,
    ) -> None:
        pass

    def __init__(self, eegdataset=None, event_id=None, rename=None) -> None:
        super().__init__(
            get_init_args(self, locals(), rename=rename, ret_dict=True),
            event_id=event_id,
        )
        self._repr.pop("rename")

        if eegdataset is None:
            self.eegdataset = {}
        elif isinstance(eegdataset, dict):
            self.eegdataset = eegdataset
        elif isinstance(eegdataset, list):
            self.eegdataset = {
                subject: egd for subject, egd in enumerate(eegdataset, start=1)
            }
        else:
            raise TypeError(f"Input type {type(eegdataset)} cannot be parsed.")

    def __setitem__(self, subject: int, eegdata: BaseData):
        """Set EEGData for subject.

        Parameters
        ----------
        subject : int, str
            Subject or session index.
        eegdata : EEGData

        """
        self.eegdataset[subject] = eegdata

    def _get_single_subject_data(self, subject: int):
        return self.eegdataset[subject]

    def get_data(self):
        """Returns the eegdata of all subjects."""
        data = {}
        for subject in self.keys():
            data[subject] = self._get_single_subject_data(subject)
        return data

    def keys(self):
        """Return the list of subjects."""
        return self.eegdataset.keys()

    def pop(self, subject: int, default=None):
        """Remove specified subject and return the corresponding eegdata.

        If the subject is not found, return the default if given; otherwise,
        raise a KeyError.
        """
        return self.eegdataset.pop(subject, default)

    def _eegdata_type(self) -> set[str]:
        type_set = set()
        for eegdata in self.eegdataset.values():
            type_set.add(type(eegdata).__name__)
        return type_set

    def __repr__(self) -> str:
        eegdataset = (
            f"Subjects={len(self.eegdataset)}"
            f", type={iterable_to_str(self._eegdata_type(), symbol='|')}"
        )
        self._repr["eegdataset"] = eegdataset
        return _format_log(self._repr)


DATA_PATH = Path(DPEEG_DIR) / "datasets"


class RawDataset(BaseDataset):
    """DPEEG Raw EEG Dataset.

    Datasets with ``mne.io.Raw`` as the base class, such as depression, use
    whether the subject suffers from depression as the label. The ``get_raw()``
    is used to obtain the raw EEG data of the subject, and ``get_data()`` is
    used to obtain data in the form of ``EEGData``. Since the raw EEGs of these
    datasets often have no labels, when calling ``get_data()``, you need to
    manually set the label through ``_set_label()`` according to different
    datasets (the label here is in units of run).
    """

    def __init__(
        self,
        repr: dict,
        subject_list: list[int],
        event_id: dict[str, int] | None = None,
        subjects: list[int] | None = None,
        tmin: float = 0.0,
        tmax: float | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        unit_factor: float = 1e6,
    ) -> None:
        super().__init__(repr, event_id)
        self.subjects = subject_list if subjects is None else subjects
        if tmax is not None and tmin >= tmax:
            raise ValueError("tmax must be greater than tmin")

        if not (set(self.subjects) <= set(subject_list)):
            raise KeyError(f"Subject must be between 1 and {subject_list[-1]}.")

        self.tmin = tmin
        self.tmax = tmax
        self.picks = picks
        self.resample = resample
        self.unit_factor = unit_factor

    def keys(self) -> list[int]:
        return self.subjects

    @abstractmethod
    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        return NotImplementedError

    def get_raw(
        self, progressbar: bool = True, verbose="ERROR"
    ) -> dict[int, dict[str, dict[str, Raw]]]:
        """Return the raw correspoonding to a list of subjects.

        The returned data is a dictionary with the following structure:

            data = {'subject_id': {'session_id': {'run_id': Raw}}}

        subjects are on top, then we have sessions, then runs.
        A session is a recording done in a single day, without removing the EEG
        cap. A session is constitued of at least one run. A run is a single
        contigous recording. Some dataset break session in multiple runs.
        """
        subjects = tqdm(
            self.subjects,
            "Load Raw",
            unit="sub",
            dynamic_ncols=True,
            disable=not progressbar,
        )

        data = {}
        for subject in subjects:
            data[subject] = self._get_single_subject_raw(subject, verbose)
        return data

    def _set_label(self, subject: int):
        return np.empty(0)

    def _get_single_subject_data(self, subject: int, verbose="ERROR"):
        data = []
        for session, runs in self._get_single_subject_raw(subject, verbose).items():  # type: ignore
            raws = []
            for run, raw in runs.items():
                if self.resample:
                    raw.resample(self.resample, verbose=verbose)

                if self.picks is None:
                    picks = mne.pick_types(raw.info, eeg=True, stim=False)
                else:
                    picks = mne.pick_channels(
                        raw.info["ch_names"], include=self.picks, ordered=True
                    )

                raws.append(
                    np.expand_dims(
                        raw.get_data(
                            picks=picks,
                            tmin=self.tmin,
                            tmax=self.tmax,
                            verbose=verbose,
                        ),
                        axis=0,
                    )
                )

            edata = np.concatenate(raws, axis=-1) * self.unit_factor
            label = self._set_label(subject)
            data.append(EEGData(edata, label))
        return MultiSessEEGData(data)

    def get_data(
        self, progressbar: bool = True, verbose="ERROR"
    ) -> dict[int, MultiSessEEGData]:
        """Return the data correspoonding to a list of subjects.

        The returned data is a dictionary with the following structure:

            data = {'subject_id' : {'session_id' : EEGData}}
        """
        subjects = tqdm(
            self.subjects,
            "Load EEGData",
            unit="sub",
            dynamic_ncols=True,
            disable=not progressbar,
        )

        data = {}
        for subject in subjects:
            data[subject] = self._get_single_subject_data(subject, verbose)
        return data


class EpochsDataset(RawDataset):
    """DPEEG Epochs EEG Dataset.

    Datasets with ``mne.Epochs`` as the base class, such as motor imagery,
    where each subject performs a different imagery task. ``get_raw()`` is used
    to obtain the raw EEG data of each subject, ``get_epochs()`` is used to
    obtain the Epochs of each subject, and ``get_data()`` is used to obtain the
    ``EEGData`` of each subject.
    """

    def __init__(
        self,
        repr: dict,
        subject_list: list[int],
        interval: list[float],
        event_id: dict[str, int],
        subjects: list[int] | None = None,
        tmin: float = 0.0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        unit_factor: float = 1e6,
    ) -> None:
        super().__init__(
            repr=repr,
            subject_list=subject_list,
            event_id=event_id,
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            picks=picks,
            resample=resample,
            unit_factor=unit_factor,
        )

        self.interval = interval
        self.baseline = baseline

    def _get_single_subject_epochs(
        self, subject: int, verbose="ERROR"
    ) -> dict[str, Epochs]:
        data = {}
        for session, runs in self._get_single_subject_raw(subject, verbose).items():  # type: ignore
            epochs = []
            for run, raw in runs.items():
                proc = self._epochs_from_raw(raw)
                if proc is None:
                    # if the run did not contain any selected event go to next
                    continue

                epochs.append(proc)
            data[session] = mne.concatenate_epochs(epochs, verbose=verbose)
        return data

    def get_epochs(
        self, progressbar: bool = True, verbose="ERROR"
    ) -> dict[int, dict[str, Epochs]]:
        """Return the epochs correspoonding to a list of subjects.

        The returned data is a dictionary with the following structure:

            data = {'subject_id' : {'session_id' : Epochs}}
        """
        subjects = tqdm(
            self.subjects,
            "Load Epochs",
            unit="sub",
            dynamic_ncols=True,
            disable=not progressbar,
        )

        data = {}
        for subject in subjects:
            data[subject] = self._get_single_subject_epochs(subject, verbose)
        return data

    def _get_single_subject_data(
        self, subject: int, verbose="ERROR"
    ) -> MultiSessEEGData:
        sessions = self._get_single_subject_epochs(subject, verbose)

        data = []
        for session, epochs in sessions.items():
            data.append(self._data_from_epochs(epochs, verbose))
        return MultiSessEEGData(data)

    def _epochs_from_raw(self, raw: Raw, verbose="ERROR") -> Epochs:
        events, event_id = self._events_from_raw(raw, verbose)

        # get interval
        tmin = self.tmin + self.interval[0]
        if self.tmax is None:
            tmax = self.interval[1]
        else:
            tmax = self.tmax + self.interval[0]

        # epoch data
        baseline = self.baseline
        if baseline is not None:
            baseline = (
                baseline[0] + self.interval[0],
                baseline[1] + self.interval[1],
            )
            bmin = baseline[0] if baseline[0] < tmin else tmin
            bmax = baseline[1] if baseline[1] > tmax else tmax
        else:
            bmin = tmin
            bmax = tmax
        epochs = mne.Epochs(
            raw,
            events,
            event_id=event_id,
            tmin=bmin,
            tmax=bmax,
            proj=False,
            baseline=baseline,
            preload=True,
            event_repeated="drop",
            verbose=verbose,
        )
        if bmin < tmin or bmax > tmax:
            epochs.crop(tmin=tmin, tmax=tmax)
        return epochs.crop(include_tmax=False)

    def _data_from_epochs(self, epochs: Epochs, verbose="ERROR") -> EEGData:
        if self.picks is None:
            picks = mne.pick_types(epochs.info, eeg=True, stim=False)
        else:
            picks = mne.pick_channels(
                epochs.info["ch_names"], include=self.picks, ordered=True
            )
        epochs.pick(picks, verbose=verbose)

        if self.resample is not None:
            epochs.resample(self.resample, verbose=verbose)

        edata = self.unit_factor * epochs.get_data(copy=False)
        label = epochs.events[:, -1]
        return EEGData(edata, label)

    def _events_from_raw(self, raw: Raw, verbose="ERROR"):
        stim_channels = mne.utils._get_stim_channel(None, raw.info, raise_error=False)
        if len(stim_channels) > 0:
            events = mne.find_events(raw, shortest_event=0, verbose=verbose)
            event_id = self.event_id
        else:
            events, event_id = mne.events_from_annotations(
                raw, event_id=self.event_id, verbose=verbose  # type: ignore
            )
            # offset = int(self.interval[0] * raw.info["sfreq"])
            # events[:, 0] -= offset  # return the original events onset

        return events, event_id



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from .base import ClsExp
from ..datasets.base import BaseDataset, EEGData, SplitEEGData
from ..transforms.base import ToEEGData
from .holdout import HoldOut
from .kfold import KFold


class LOSO(ClsExp):
    """Leave-One-Subject-Out cross validation experiment.

    Suppose that a dataset has :math:`N_s` subjects. In the LOSO_HoldOut
    experiment, the data of each subject was used as once as the test set, and
    the data of the remaining :math:`N_s - 1` subjects formed the training set.
    The average classification accuracy of the model was then calculated from
    the :math:`N_s` analyses.
    """

    _is_eeg_dataset = True  # Input data type must be a eeg dataset

    def _process_sub_dataset(self, dataset: BaseDataset, subject: int):
        """Merge eegdata according to LOSO experiment requirements."""
        if len(dataset) < 2:
            raise RuntimeError("Dataset has at least 2 subjects in LOSO.")

        self.logger.info(f"[Leave Subject {subject} Out]")
        train_egd = EEGData()
        for sub, egd in dataset.items():
            egd = ToEEGData()(egd, verbose=False)
            if sub == subject:
                test_egd = egd
            else:
                train_egd.append(egd)

        return SplitEEGData(train_egd, test_egd)


class LOSO_HoldOut(HoldOut, LOSO):
    r"""Leave-One-Subject-Out Holdout cross validation experiment.

    See Also
    --------
    :class:`.HoldOut`
    """

    pass


class LOSO_KFold(KFold, LOSO):
    """Leave-One-Subject-Out K-Fold cross validation experiment.

    See Also
    --------
    :class:`.KFold`
    """

    def _process_sub_dataset(self, dataset: BaseDataset, subject: int):
        if self.isolate_testset == False:
            self.isolate_testset = True
            self.logger.warning(
                "The `isolate_testset` is ignored in LOSO_KFold experiment."
            )
        return super()._process_sub_dataset(dataset, subject)



import torch
import torch.nn.functional as F
import torch.nn as nn
from torch import Tensor, fill
from einops import rearrange
from einops.layers.torch import Rearrange

from ..tools.docs import fill_doc


__all__ = ["EEGConformer"]


class PatchEmbedding(nn.Module):
    def __init__(self, nCh, emb_size=40):
        super().__init__()

        self.shallownet = nn.Sequential(
            nn.Conv2d(1, 40, (1, 25), (1, 1)),
            nn.Conv2d(40, 40, (nCh, 1), (1, 1)),
            nn.BatchNorm2d(40),
            nn.ELU(),
            nn.AvgPool2d((1, 75), (1, 15)),
            nn.Dropout(0.5),
        )

        self.projection = nn.Sequential(
            nn.Conv2d(40, emb_size, (1, 1), stride=(1, 1)),
            Rearrange("b e (h) (w) -> b (h w) e"),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.shallownet(x)
        x = self.projection(x)
        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, emb_size, num_heads, dropout):
        super().__init__()
        self.emb_size = emb_size
        self.num_heads = num_heads
        self.keys = nn.Linear(emb_size, emb_size)
        self.queries = nn.Linear(emb_size, emb_size)
        self.values = nn.Linear(emb_size, emb_size)
        self.att_drop = nn.Dropout(dropout)
        self.projection = nn.Linear(emb_size, emb_size)

    def forward(self, x: Tensor) -> Tensor:
        queries = rearrange(self.queries(x), "b n (h d) -> b h n d", h=self.num_heads)
        keys = rearrange(self.keys(x), "b n (h d) -> b h n d", h=self.num_heads)
        values = rearrange(self.values(x), "b n (h d) -> b h n d", h=self.num_heads)
        energy = torch.einsum("bhqd, bhkd -> bhqk", queries, keys)

        scaling = self.emb_size ** (1 / 2)
        att = F.softmax(energy / scaling, dim=-1)
        att = self.att_drop(att)
        out = torch.einsum("bhal, bhlv -> bhav ", att, values)
        out = rearrange(out, "b h n d -> b n (h d)")
        out = self.projection(out)
        return out


class ResidualAdd(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        res = x
        x = self.fn(x, **kwargs)
        x += res
        return x


class FeedForwardBlock(nn.Sequential):
    def __init__(self, emb_size, expansion, drop_p):
        super().__init__(
            nn.Linear(emb_size, expansion * emb_size),
            nn.GELU(),
            nn.Dropout(drop_p),
            nn.Linear(expansion * emb_size, emb_size),
        )


class TransformerEncoderBlock(nn.Sequential):
    def __init__(
        self,
        emb_size,
        num_heads=10,
        drop_p=0.5,
        forward_expansion=4,
        forward_drop_p=0.5,
    ):
        super().__init__(
            ResidualAdd(
                nn.Sequential(
                    nn.LayerNorm(emb_size),
                    MultiHeadAttention(emb_size, num_heads, drop_p),
                    nn.Dropout(drop_p),
                )
            ),
            ResidualAdd(
                nn.Sequential(
                    nn.LayerNorm(emb_size),
                    FeedForwardBlock(
                        emb_size, expansion=forward_expansion, drop_p=forward_drop_p
                    ),
                    nn.Dropout(drop_p),
                )
            ),
        )


class TransformerEncoder(nn.Sequential):
    def __init__(
        self, depth, emb_size, num_heads, drop_p, forward_expansion, forward_drop_p
    ):
        super().__init__(
            *[
                TransformerEncoderBlock(
                    emb_size, num_heads, drop_p, forward_expansion, forward_drop_p
                )
                for _ in range(depth)
            ]
        )


class ClassificationHead(nn.Sequential):
    def __init__(self, in_features, nCls):
        super().__init__(
            nn.Flatten(),
            nn.Linear(in_features, 256),
            nn.ELU(),
            nn.Dropout(0.5),
            nn.Linear(256, 32),
            nn.ELU(),
            nn.Dropout(0.3),
            nn.Linear(32, nCls),
            nn.LogSoftmax(dim=1),
        )


@fill_doc
class EEGConformer(nn.Module):
    """EEG Conformer: Convolutional Transformer for EEG Decoding and
    Visualization (EEG Conformer).

    EEG Conformer [1]_ is proposed to encapsulate local and global features in
    a unified EEG classification framework. The architecture comprises three
    components: a convolution module, a self-attention module, and a fully-
    connected classifier. In the convolution module, taking the raw two-
    dimensional EEG trials as the input, temporal and spatial convolutional
    layers are applied along the time dimension and electrode channel
    dimensions, respectively. Then, an average pooling layer is utilized to
    suppress noise interference while improving generalization. Secondly, the
    spatial-temporal representation obtained by the convolution module is fed
    into the selfattention module. The self-attention module further extracts
    the long-term temporal features by measuring the global correlations
    between different time positions in the feature maps. Finally, a compact
    classifier consisting of several fullyconnected layers is adopted to output
    the decoding results.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    emb_size : int
        Embedding layer size.
    depth : int
        Depth of transformer encoder.
    num_heads : int
        Number of multi-head attention.
    drop_p : float
        Dropout rate of transformer encoder.
    forward_expansion : int
        The expansion factor of the fully connected feed-forward layer.
    forward_drop_p : float
        Dropout rate of fully connected feed-forward layer.

    References
    ----------
    .. [1] Y. Song, Q. Zheng, B. Liu, and X. Gao, “EEG conformer: Convolutional
        transformer for EEG decoding and visualization,” IEEE Transactions on
        Neural Systems and Rehabilitation Engineering, vol. 31, pp. 710–719,
        2022.
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        emb_size: int = 40,
        depth: int = 6,
        num_heads: int = 10,
        drop_p: float = 0.5,
        forward_expansion: int = 4,
        forward_drop_p: float = 0.5,
    ) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime
        self.patch_embedding = PatchEmbedding(nCh, emb_size)
        self.transformer_encoder = TransformerEncoder(
            depth, emb_size, num_heads, drop_p, forward_expansion, forward_drop_p
        )
        in_freatures = self._forward_transformer().size(1)
        self.classification_head = ClassificationHead(in_freatures, nCls)

    def _forward_transformer(self) -> torch.Tensor:
        x = torch.randn(1, 1, self.nCh, self.nTime)
        x = self.patch_embedding(x)
        x = self.transformer_encoder(x)
        return x.flatten(start_dim=1, end_dim=-1)

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, 1, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        out = self.patch_embedding(x)
        out = self.transformer_encoder(out)
        return self.classification_head(out)



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from .base import ClsExp
from ..trainer.classifier import BaseClassifier
from ..transforms.base import SplitTrainTest
from ..utils import get_init_args
from ..tools.docs import fill_doc


@fill_doc
class HoldOut(ClsExp):
    """Holdout cross validation experiment.

    Validate the performance of the model on unseen data using holdout cross
    validation. Only one training and testing is required, so it is fast and
    suitable for large-scale datasets and fast model evaluation.

    Parameters
    ----------
    %(trainer)s
    %(out_folder)s
    %(timestamp)s
    """

    def __init__(
        self,
        trainer: BaseClassifier,
        out_folder: str | None = None,
        timestamp: bool = True,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(
            get_init_args(self, locals(), ret_dict=True),
            trainer=trainer,
            out_folder=out_folder,
            timestamp=timestamp,
            verbose=verbose,
        )

    def _run_sub_classifier(self, eegdata, sub_folder):
        result = self.trainer.fit(
            trainset=eegdata["train"],
            testset=eegdata["test"],
            log_dir=sub_folder,
        )
        return result

    def _run_sub_classifier_two_stage(self, eegdata, sub_folder):
        split_eegdata = SplitTrainTest()(eegdata["train"])
        train_set = split_eegdata["train"]
        valid_set = split_eegdata["test"]
        test_set = eegdata["test"]

        result = self.trainer.fit(
            trainset=train_set,
            validset=valid_set,
            testset=test_set,
            log_dir=sub_folder,
        )
        return result

    def _run_sub(self, eegdata, sub_folder):
        self.logger.info(f"\n# ---------- {sub_folder.name} ---------- #")
        result = self._run_sub_func(self._trans_eegdata(eegdata), sub_folder)
        return (
            result["test"]["acc"],
            result["test"]["preds"],
            result["test"]["target"],
            result,
        )



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import sys
import json
from pathlib import Path
from collections.abc import Iterable
from typing import TypeVar

import numpy as np
from numpy.lib import format
from mne.utils import verbose, logger
from tqdm import tqdm

from .. import __version__
from ..utils import iterable_to_str
from .base import BaseData, BaseDataset, EEGData, MultiSessEEGData, SplitEEGData


CURRENT_MODULE = sys.modules[__name__]
T = TypeVar("T")


@verbose
def check_inter_and_compl(
    r_set: Iterable[T], q_set: Iterable[T], strict: bool = True, verbose=None
) -> tuple[set[T], set[T]]:
    """Check the intersection and complement of two sets.

    Check whether the search set and query set meet the requirements and return
    the corresponding intersection and complement.

    Parameters
    ----------
    r_set : Iterable of Any
        Retrieve a collection.
    q_set : Iterable of Any
        Query a collection.
    strict : bool
        By default, an error is raised if any element in the retrieved set does
        not exist in the query set. If `False`, the intersection is allowed to
        be empty.

    Returns
    -------
    inter : set
        Retrieves the intersection of a set and a query set.
    compl : set
        Retrieves the complement of a set with respect to a query set.

    Examples
    --------
    >>> r_set, q_set = [1, 2, 3], [2, 3, 4]
    >>> inter, compl = check_inter_and_compl(r_set, q_set)
    Cannot find 1, only use 2, 3.
    >>> inter, compl
    ({2, 3}, {1})
    """
    r_set, q_set = set(r_set), set(q_set)

    inter = r_set & q_set
    if len(inter) == 0:
        if strict:
            raise KeyError(
                f"Cannot find {iterable_to_str(r_set)}, "
                f"data only contains {iterable_to_str(q_set)}."
            )
        else:
            logger.warning(f"Cannot find {iterable_to_str(r_set)}.")

    compl = r_set - q_set
    if len(compl) != 0:
        logger.warning(
            f"Cannot find {iterable_to_str(compl)}, "
            f"only use {iterable_to_str(inter)}."
        )

    return inter, compl


def save(
    file: str | Path,
    eegdata: BaseData,
    compress: bool = False,
):
    """Archive eegdata into a file in '.egd' format.

    Parameters
    ----------
    file : str, Path
        Filename where the eegdata will be saved, a '.egd' extension will be
        appended to the filename if it does not already have one.
    eegdata : :ref:`eeg_data`
        An eegdata to be saved.
    compress : bool
        If ``True``, use standard ZIP compression.

    Returns
    -------
    path : Path
        File save position.
    """
    import zipfile
    import json

    if compress:
        compression = zipfile.ZIP_DEFLATED
    else:
        compression = zipfile.ZIP_STORED

    path = Path(file).resolve().with_suffix(".egd")
    with zipfile.ZipFile(path, mode="w", compression=compression) as zipf:
        eegdata_info = {"version": __version__, "type": type(eegdata).__name__}
        zipf.writestr("eegdata_info.json", json.dumps(eegdata_info, indent=4))

        for val, key in eegdata._datas():
            for k, v in val.items():
                fname = k + ".npy" if key is None else Path(key) / (k + ".npy")
                with zipf.open(str(fname), "w", force_zip64=True) as fid:
                    format.write_array(fid, v, allow_pickle=True)

    return path


def save_dataset(
    folder: str | Path,
    dataset: BaseDataset,
    compress: bool = False,
    name_folder: bool = True,
    progressbar: bool = True,
):
    """Save the eegdata dataset to a folder.

    The eegdata in the dataset is stored separately by subject, and a `.json`
    formatted file describing the basic information of the dataset is saved.
    Use :class:`dpeeg.datasets.LoadDataset` to read the saved dataset.

    Parameters
    ----------
    folder : str, Path
        The folder location where the dataset is saved.
    dataset : eegdata dataset
        An eegdata dataset to be saved.
    compress : bool
        If `True`, each subject's eegdata is saved using standard ZIP
        compression.
    name_folder : bool
        If `True`, a subfolder with the name of the dataset is created under
        the folder `folder` and saved.
    progressbar : bool
        Whether to show the progress bar.

    Returns
    -------
    path : Path
        Dataset save folder.
    """
    path = Path(folder).resolve()
    if name_folder:
        path = path / dataset._repr["_obj_name"]

    path.mkdir(parents=True, exist_ok=True)
    if any(path.iterdir()):
        raise FileExistsError(f"'{str(path)}' is not a empty folder.")

    dataset_info = {
        "version": __version__,
        "name": dataset._repr["_obj_name"],
        "event_id": dataset._repr["event_id"],
    }
    with open(path / "dataset_info.json", "w") as filer:
        json.dump(dataset_info, filer, indent=4)

    subjects = tqdm(
        dataset.items(),
        "Save dataset",
        total=len(dataset),
        unit="sub",
        dynamic_ncols=True,
        disable=not progressbar,
    )

    for sub, sub_data in subjects:
        save(path / f"sub_{sub}", sub_data, compress=compress)

    return path


def _load_EEGData(zipf, filelist) -> EEGData:
    data = {}
    for file in filelist:
        with zipf.open(file, "r") as fid:
            data[Path(file).stem] = format.read_array(fid, allow_pickle=True)
    return EEGData(**data)


def _load_two_wrapping_eegdata(zipf, filelist):
    tmp = {}
    for file in filelist:
        sess, fname = Path(file).parts

        with zipf.open(file, "r") as fid:
            _data = format.read_array(fid, allow_pickle=True)
        tmp.setdefault(sess, {}).update({Path(fname).stem: _data})

    data = {}
    for sess, _data in tmp.items():
        data[sess] = EEGData(**_data)

    return data


def _load_MultiSessEEGData(zipf, filelist) -> MultiSessEEGData:
    return MultiSessEEGData(_load_two_wrapping_eegdata(zipf, filelist))


def _load_SplitEEGData(zipf, filelist) -> SplitEEGData:
    return SplitEEGData(**_load_two_wrapping_eegdata(zipf, filelist))


def load(file: str | Path):
    """Load eegdata from `.egd` file."""
    import zipfile
    import json

    path = Path(file).resolve(strict=True)
    with zipfile.ZipFile(path, mode="r") as zipf:
        filelist = zipf.namelist()
        eegdata_info = "eegdata_info.json"

        if eegdata_info not in filelist:
            raise FileNotFoundError(f"eegdata info not found in {path} file.")
        with zipf.open(eegdata_info, "r") as info:
            data = json.loads(info.read())
            eegdata_type = data["type"]
        filelist.remove(eegdata_info)

        return getattr(CURRENT_MODULE, f"_load_{eegdata_type}")(zipf, filelist)



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from pathlib import Path
from abc import ABC, abstractmethod
from inspect import getmro

import torch
from torchmetrics.aggregation import CatMetric, MeanMetric

import dpeeg
from ..datasets.base import SplitEEGData, BaseDataset, BaseData
from ..transforms.base import Transforms
from ..trainer.base import Trainer
from ..trainer.classifier import BaseClassifier
from ..tools import Logger, Timer, Filer
from ..utils import DPEEG_DIR, iterable_to_str, _format_log, _format_log_kv


class Experiment(ABC):
    """Experiment base class.

    Parameters
    ----------
    repr : str
        Basic information on the experiment method.
    trainer : Trainer
        Trainer used for training module on dataset.
    out_folder : str, optional
        Store all experimental results in a folder named with the model class
        name in the specified folder. Default is
        '~/dpeeg/out/model/exp/dataset/timestamp'.
    timestamp : bool
        Output folders are timestamped.
    verbose : int, str
        The log level of console. Default is INFO. Mainly used for debugging.

    Notes
    -----
    The training results of all models for each subject will be saved under the
    ``out_folder`` directory.
    """

    # Special control properties
    _is_eeg_dataset = False  # Input data type must be a eeg dataset

    def __init__(
        self,
        repr: dict,
        trainer: Trainer,
        out_folder: str | None = None,
        timestamp: bool = True,
        verbose: int | str = "INFO",
    ) -> None:
        repr.pop("trainer")
        self._repr = repr
        self.trainer = trainer
        self.timestamp = timestamp
        self.verbose = verbose

        # create logger and timer
        self.logger = Logger("dpeeg_exp", clevel=verbose)
        self.timer = Timer()

        # set output folder
        net = trainer.model.__class__.__name__
        exp = self.__class__.__name__
        self.out_folder = (
            Path(out_folder).absolute().joinpath(net, exp)
            if out_folder
            else Path(DPEEG_DIR).joinpath("out", net, exp)
        )

    @abstractmethod
    def _run(self) -> dict:
        """Reconstruct different experimental processes according to different
        experimental designs.
        """
        pass

    def run(
        self,
        dataset: BaseDataset | BaseData,
        transforms: Transforms | None = None,
        dataset_name: str | None = None,
        desc: str | None = None,
    ) -> dict:
        """Train models separately for each subject.

        This function will internally call the `_run_sub` function for each
        subject, and save the final results together.

        Parameters
        ----------
        dataset : :ref:`eeg_data` or :ref:`eeg_dataset`
            The data or dataset used for the experimental test.
        transforms : Transforms, optional
            Apply pre-transforms on dataset. Transformations will be apply
            during the experiment on each subject's dataset. The rationable
            behind this method lies in deferring data manipulation, especially
            for certain transformations that could potentially enlarge the
            dataset's memory footprint. This delay allows for the execution of
            data manipulation after subject-independent experiment have
            concatenated the relevant data (Time for Space) or the experiment
            subject are ready, mitigating the risk of memory overflow.
        dataset_name : str, optional
            The dataset name to use. If ``None``, The default name of the
            dataset will be used as the folder to save experimental results.
        desc : str, optional
            Add a short description to the current experiment.

        Returns
        -------
        Return a dict of all subjects and corresponding experimental results.
        """

        if isinstance(dataset, BaseDataset):
            self.data_folder = (
                self.out_folder / dataset_name
                if dataset_name
                else self.out_folder / dataset._repr["_obj_name"]
            )

        elif isinstance(dataset, BaseData):
            if self._is_eeg_dataset:
                raise RuntimeError("Experiments only support eeg dataset.")
            else:
                self.data_folder = (
                    self.out_folder / dataset_name if dataset_name else self.out_folder
                )

        else:
            raise TypeError("Input dataset must be eeg data or dataset.")

        if self.timestamp:
            self.data_folder = self.data_folder / Timer.cdate()
        self.data_folder.mkdir(parents=True, exist_ok=False)
        self.logger.info(f"Results saved in `{self.data_folder}`")

        # Experimental Resources
        self.dataset = dataset
        self.transforms = transforms

        self.filer = Filer(self.data_folder / "summary.txt")
        self.filer.write(f"[Start Time]: {self.timer.ctime()}\n")
        self.filer.write(f"[DPEEG Version]: {dpeeg.__version__}\n")
        self.filer.write(f"[Description]: {desc}\n")
        self.filer.write(str(dataset) + "\n")
        self.filer.write(_format_log_kv("Transforms", transforms) + "\n")
        self.filer.write(_format_log_kv("Trainer", self.trainer) + "\n")
        self.filer.write(str(self) + "\n")

        self.timer.start()
        self.logger.info("=" * 50)

        # Start the experiment for all subjects.
        results = self._run()

        h, m, s = self.timer.stop()
        torch.save(results, self.data_folder / f"results.pt")
        self.logger.info(f"\n[Experiment Finished] - [Cost Time = {h}H:{m}M:{s:.2f}S]")
        self.logger.info("=" * 50)

        return results

    def __repr__(self) -> str:
        return _format_log(self._repr)


class ClsExp(Experiment, ABC):
    """Base class for classification experiments."""

    @abstractmethod
    def _run_sub_classifier(self, *args, **kwargs) -> dict:
        pass

    @abstractmethod
    def _run_sub_classifier_two_stage(self, *args, **kwargs) -> dict:
        pass

    def __init__(
        self,
        repr: dict,
        trainer: BaseClassifier,
        out_folder: str | None = None,
        timestamp: bool = True,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(
            repr=repr,
            trainer=trainer,
            out_folder=out_folder,
            timestamp=timestamp,
            verbose=verbose,
        )

        trainer_type = [base.__name__ for base in getmro(type(trainer))]
        trainer_list = {
            "Classifier": self._run_sub_classifier,
            "ClassifierTwoStage": self._run_sub_classifier_two_stage,
        }
        inter = set(trainer_list) & set(trainer_type)
        if len(inter) == 0:
            raise TypeError(
                f"Trainer type {trainer_type} is not supported, "
                f"only {iterable_to_str(trainer_list.keys())} are supported."
            )
        else:
            self._run_sub_func = trainer_list[inter.pop()]

    def _trans_eegdata(self, eegdata: BaseData) -> SplitEEGData:
        """Apply pre-transforms on eegdata.

        Raises
        ------
        TypeError
            If the eegdata is not split after transformed.
        """
        if self.transforms is not None:
            eegdata = self.transforms(eegdata)

        if not isinstance(eegdata, SplitEEGData):
            raise TypeError("The eegdata is not split.")

        return eegdata

    def _process_sub_dataset(self, dataset: BaseDataset, subject: int):
        """Preprocess each subject's dataset.

        Different preprocessing operations are performed on the dataset accord-
        ing to different experimental requirement. By default, eegdata for each
        subject in the dataset is returned.
        """
        return dataset[subject]

    @abstractmethod
    def _run_sub(self, eegdata: BaseData, sub_folder: Path) -> dict:
        """Train a model on the specified subject data.

        This function will be called by `_run` function to conduct experiments
        on the data of each individual subject. Reconstruct the model training
        process according to different experimental requirements.

        Parameters
        ----------
        subject : int
            Subject of the experiment. Create a subdirectory of `out_folder` to
            store all yield results during subject training.
        eegdata : eegdata
            Subject eegdata. Adjust the subject eegdata according to different
            experiments.

        Returns
        -------
        result : dict
            Contains the `acc`, `preds`, `target` and detailed results.
        """
        pass

    def _run(self) -> dict:
        if isinstance(self.dataset, BaseDataset):
            return self._run_eegdata_set(self.dataset)
        else:
            return self._run_eegdata(self.dataset)

    def _run_eegdata(self, dataset: BaseData) -> dict:
        acc, preds, target, detail = self._run_sub(dataset, self.data_folder)
        return {"acc": acc, "preds": preds, "target": target, "detail": detail}

    def _run_eegdata_set(self, dataset: BaseDataset) -> dict:
        result = {}
        acc_metric = MeanMetric()
        preds_metric = CatMetric()
        target_metric = CatMetric()

        for subject in dataset.keys():
            self.logger.info(f"\n[Subject-{subject} Training ...]")
            self.logger.info("-" * 50)

            eegdata = self._process_sub_dataset(dataset, subject)
            sub_folder = self.data_folder / f"sub{subject}"
            sub_folder.mkdir(parents=True, exist_ok=False)
            acc, preds, target, subject_result = self._run_sub(eegdata, sub_folder)  # type: ignore
            result[f"subject_{subject}"] = subject_result

            acc_metric.update(acc)
            preds_metric.update(preds)
            target_metric.update(target)

            self.filer.write(f"Subject_{subject} Acc = {acc*100:.2f}%\n")

        acc = acc_metric.compute()
        self.filer.write(f"Model Acc = {acc*100:.2f}%\n")

        self.logger.info("-" * 50)
        self.logger.info(f"[Model Acc = {acc*100:.2f}%]")

        result.update(
            {
                "acc": acc,
                "preds": preds_metric.compute(),
                "target": target_metric.compute(),
            }
        )
        return result



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from abc import abstractmethod

from mne.utils import verbose, logger

from ..datasets.base import _BaseData, EEGData, SplitEEGData
from .base import Transforms
from ..utils import DPEEG_SEED, get_init_args
from ..tools.docs import fill_doc
from .functions import (
    segmentation_and_reconstruction_time,
    slide_win,
    crop,
    gaussian_noise_time,
)


class Augmentation(Transforms):
    """Data augmentation base class.

    Augment the data, with default augmentation applied to the ``edata`` and
    ``label``. Please verify the validity of the data.
    """

    def __init__(
        self, repr: str | None = None, only_train: bool = True, strict: bool = True
    ) -> None:
        super().__init__(repr)
        self.only_train = only_train
        self.strict = strict

    def _apply(self, eegdata: _BaseData) -> _BaseData:
        if not isinstance(eegdata, SplitEEGData):
            if self.strict:
                raise TypeError(
                    "The input must have been split, or `strict` is set to False."
                )
            else:
                for egd, _ in eegdata._datas():
                    self._apply_aug(egd, "None")
        else:
            for egd, mode in eegdata._datas():
                if (mode != "train") and self.only_train:
                    continue
                self._apply_aug(egd, mode)

        return eegdata

    @abstractmethod
    def _apply_aug(self, egd: EEGData, mode: str):
        pass


@fill_doc
class SegRecTime(Augmentation):
    """Segmentation and reorganization in the time domain.

    The S&R process involves segmenting the original eeg signals based on class
    along the temporal dimension, followed by randomly splicing them back [1]_.
    By default, augmentation is performed on ``edata`` and ``label``. Ensure
    the availability of the data.

    Parameters
    ----------
    samples : int
        The number of consecutive samples to segment the data. eg, 125 for
        250Hz data is segmented by 0.5s.
    multiply : float
        Data expansion multiple of relative metadata, 1 means doubled.
    %(aug_only_train)s
    %(aug_strict)s
    seed : int
        Seed to be used to instantiate numpy random number generator instance.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    References
    ----------
    .. [1] F. Lotte, “Signal processing approaches to minimize or suppress
        calibration time in oscillatory activity-based brain-computer
        interfaces,” Proceedings of the IEEE, vol. 103, no. 6,
        pp. 871-890, 2015.

    Notes
    -----
    Data augmentation is only applied to the `edata` and `label` within the
    eegdata, with other values remaining unchanged. If there are derived values
    based on the `edata`, attention should be paid to the order of
    transformations.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> split_eegdata = dpeeg.SplitEEGData(eegdata.copy(), eegdata.copy())
    >>> transforms.SegRecTime(2, 3)(split_eegdata, verbose=False)
    Train: [edata=(64, 3, 10), label=(64,)]
    Test : [edata=(16, 3, 10), label=(16,)]
    """

    def __init__(
        self,
        samples: int,
        multiply: float = 1.0,
        only_train: bool = True,
        strict: bool = True,
        seed: int = DPEEG_SEED,
    ) -> None:
        super().__init__(
            get_init_args(self, locals(), format="rp"),
            only_train=only_train,
            strict=strict,
        )
        self.samples = samples
        self.multiply = multiply
        self.seed = seed

    def _apply_aug(self, egd: EEGData, mode: str):
        egd["edata"], egd["label"] = segmentation_and_reconstruction_time(
            data=egd["edata"],
            label=egd["label"],
            samples=self.samples,
            multiply=self.multiply,
            seed=self.seed,
        )


class SlideWinAug(Augmentation):
    """Sliding window data augmentation.

    Data augmentation based on sliding windows will apply sliding windows to
    the training set and crop the corresponding time windows in the test set.
    By default, augmentation is performed on ``edata`` and ``label``. Ensure
    the availability of the data.

    Parameters
    ----------
    win : int
        The size of the sliding window.
    overlap : int
        The amount of overlap between adjacent sliding windows.
    tmin : int
        Start time of selection in sampling points.
    tmax : int, optional
        End time of selection in sampling points. The default is to use the
        window length from the start time.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.

    Notes
    -----
    Data augmentation is only applied to the `edata` and `label` within the
    eegdata, with other values remaining unchanged. If there are derived values
    based on the `edata`, attention should be paid to the order of
    transformations.

    Examples
    --------
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> split_eegdata = dpeeg.SplitEEGData(eegdata.copy(), eegdata.copy())
    >>> transforms.SlideWinAug(2)(split_eegdata, verbose=False)
    Train: [edata=(80, 3, 2), label=(80,)]
    Test : [edata=(16, 3, 2), label=(16,)]
    """

    def __init__(
        self,
        win: int,
        overlap: int = 0,
        tmin: int = 0,
        tmax: int | None = None,
    ) -> None:
        super().__init__(
            get_init_args(self, locals(), format="rp"),
            only_train=False,
            strict=True,
        )
        self.win = win
        self.overlap = overlap
        self.tmin = tmin
        self.tmax = tmin + win if tmax is None else tmax

    def _apply_aug(self, egd: EEGData, mode: str):
        if mode == "train":
            egd["edata"], egd["label"] = slide_win(
                data=egd["edata"],
                win=self.win,
                overlap=self.overlap,
                label=egd["label"],
            )
        else:
            egd["edata"] = crop(
                data=egd["edata"],
                tmin=self.tmin,
                tmax=self.tmax,
                include_tmax=False,
            )


@fill_doc
class GaussTime(Augmentation):
    """Randomly add white noise to all channels.

    Gaussian white noise with a mean of 0 is directly added to the raw EEG
    signal as the generated new data [1]_. By default, augmentation is
    performed on ``edata`` and ``label``. Ensure the availability of the data.

    Parameters
    ----------
    std : float
        Standard deviation to use for the additive noise.
    %(aug_only_train)s
    %(aug_strict)s
    seed : int
        Seed to be used to instantiate numpy random number generator instance.

    References
    ----------
    .. [1] Wang, F., Zhong, S. H., Peng, J., Jiang, J., & Liu, Y. (2018). Data
       augmentation for eeg-based emotion recognition with deep convolutional
       neural networks. In International Conference on Multimedia Modeling
       (pp. 82-93).
    """

    def __init__(
        self,
        std: float,
        only_train: bool = True,
        strict: bool = True,
        seed: int = DPEEG_SEED,
    ) -> None:
        super().__init__(
            get_init_args(self, locals(), format="rp"),
            only_train=only_train,
            strict=strict,
        )
        self.std = std
        self.seed = seed

    def _apply_aug(self, egd: EEGData, mode: str):
        egd["edata"], egd["label"] = gaussian_noise_time(
            data=egd["edata"],
            label=egd["label"],
            mean=0,
            std=self.std,
            seed=self.seed,
        )



import numpy as np
from scipy.io import loadmat
from numpy import ndarray
from mne.channels import make_standard_montage
from mne.io import RawArray
from mne import create_info

from .base import EpochsDataset, DATA_PATH
from .download import data_dl
from ..utils import get_init_args
from ..tools.docs import fill_doc


URL = "http://bnci-horizon-2020.eu/database/data-sets/"


def standardize_keys(d):
    master_list = [
        ["both feet", "feet"],
        ["left hand", "left_hand"],
        ["right hand", "right_hand"],
        ["FEET", "feet"],
        ["HAND", "right_hand"],
        ["NAV", "navigation"],
        ["SUB", "subtraction"],
        ["WORD", "word_ass"],
    ]
    for old, new in master_list:
        if old in d.keys():
            d[new] = d.pop(old)


def _convert_run(run, ch_names, ch_types, verbose):
    """Convert one run to raw."""
    event_id = {}
    n_chan = run.X.shape[1]
    montage = make_standard_montage("standard_1005")
    eeg_data = 1e-6 * run.X
    sfreq = run.fs

    if not ch_names:
        ch_names = [f"EEG{ch}" for ch in range(1, n_chan + 1)]
        montage = None

    if not ch_types:
        ch_types = ["eeg"] * n_chan

    trigger = np.zeros((len(eeg_data), 1))
    # some runs does not contains trials i.e baseline runs
    if len(run.trial) > 0:
        trigger[run.trial - 1, 0] = run.y
    else:
        return None, None

    eeg_data = np.c_[eeg_data, trigger]
    ch_names = ch_names + ["stim"]
    ch_types = ch_types + ["stim"]
    event_id = {ev: (ii + 1) for ii, ev in enumerate(run.classes)}
    info = create_info(ch_names=ch_names, ch_types=ch_types, sfreq=sfreq)
    raw = RawArray(data=eeg_data.T, info=info, verbose=verbose)
    raw.set_montage(montage)
    return raw, event_id


def _convert_mi(filename, ch_names, ch_types, verbose):
    runs = []
    event_id = {}
    data = loadmat(filename, struct_as_record=False, squeeze_me=True)

    if isinstance(data["data"], ndarray):
        run_array = data["data"]
    else:
        run_array = [data["data"]]

    for run in run_array:
        raw, evd = _convert_run(run, ch_names, ch_types, verbose)
        if raw is None:
            continue
        runs.append(raw)
        event_id.update(evd)  # type: ignore
        # change labels to match rest
    standardize_keys(event_id)
    return runs, event_id


@fill_doc
class BCICIV2A(EpochsDataset):
    """Dataset IIA from BCI Competition IV.

    .. admonition:: Dataset summary

        ====== ====== ====== ===== ======= ======
        Subj   Chan   Time   Cls   Freq    Sess
        ====== ====== ====== ===== ======= ======
        9      22     4 s    4     250 Hz  2
        ====== ====== ====== ===== ======= ======

    This data set consists of EEG data from 9 subjects of a study published in
    [1]_.  The cue-based BCI paradigm consisted of four different motor imagery
    tasks, namely the imagination of movement of the left hand (class 1), right
    hand (class 2), both feet (class 3), and tongue (class 4). Two sessions on
    different days were recorded for each subject.  Each session is comprised
    of 6 runs separated by short breaks. One run consists of 48 trials (12 for
    each of the four possible classes), yielding a total of 288 trials per
    session.

    The subjects were sitting in a comfortable armchair in front of a computer
    screen.  At the beginning of a trial ( t = 0 s), a fixation cross appeared
    on the black screen.  In addition, a short acoustic warning tone was
    presented.  After two seconds ( t = 2 s), a cue in the form of an arrow
    pointing either to the left, right, down or up (corresponding to one of the
    four classes left hand, right hand, foot or tongue) appeared and stayed on
    the screen for 1.25 s.  This prompted the subjects to perform the desired
    motor imagery task.  No feedback was provided.  The subjects were ask to
    carry out the motor imagery task until the fixation cross disappeared from
    the screen at t = 6 s.

    Twenty-two Ag/AgCl electrodes (with inter-electrode distances of 3.5 cm)
    were used to record the EEG; the montage is shown in Figure 3 left. All
    signals were recorded monopolarly with the left mastoid serving as
    reference and the right mastoid as ground. The signals were sampled with.
    250 Hz and bandpass-filtered between 0.5 Hz and 100 Hz. The sensitivity of
    the amplifier was set to 100 μV . An additional 50 Hz notch filter was
    enabled to suppress line noise

    References
    ----------
    .. [1] Tangermann, M., Müller, K.R., Aertsen, A., Birbaumer, N., Braun, C.,
        Brunner, C., Leeb, R., Mehring, C., Miller, K.J., Mueller-Putz, G. and
        Nolte, G., 2012. Review of the BCI competition IV.
        Frontiers in neuroscience, 6, p.55.

    Parameters
    ----------
    %(subjects)s
    %(epochs_tmin_tmax)s
    %(baseline)s
    %(picks)s
    %(resample)s
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        rename: str | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 10)),
            interval=[2.0, 6.0],
            event_id={"left_hand": 1, "right_hand": 2, "feet": 3, "tongue": 4},
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            baseline=baseline,
            picks=picks,
            resample=resample,
        )
        # fmt: off
        self._ch_names = [
            "Fz", "FC3", "FC1", "FCz", "FC2", "FC4", "C5", "C3", "C1", "Cz", "C2",
            "C4", "C6", "CP3", "CP1", "CPz", "CP2", "CP4", "P1", "Pz", "P2", "POz",
            "EOG1", "EOG2", "EOG3",
        ]
        # fmt: on
        self._ch_types = ["eeg"] * 22 + ["eog"] * 3
        self._data_url = f"{URL}001-2014/"
        self._data_path = DATA_PATH / "bciciv2a"

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        sessions = {}
        for i, r in enumerate(["T", "E"], start=1):
            filename = data_dl(
                f"{self._data_url}A{subject:02d}{r}.mat",
                self._data_path,
                force_update=False,
            )
            runs, _ = _convert_mi(
                filename, self._ch_names, self._ch_types, verbose=verbose
            )
            sessions[f"session_{i}"] = {
                f"run_{ii}": run for ii, run in enumerate(runs, start=1)
            }
        return sessions


@fill_doc
class BCICIV2B(EpochsDataset):
    """Dataset IIB from BCI Competition IV.

    .. admonition:: Dataset summary

        ====== ====== ======= ===== ======== ======
        Subj   Chan   Time    Cls   Freq     Sess
        ====== ====== ======= ===== ======== ======
        9      3      4.5 s   2     250 Hz   5
        ====== ====== ======= ===== ======== ======


    This data set consists of EEG data from 9 subjects of a study published in
    [1]_. The subjects were right-handed, had normal or corrected-to-normal
    vision and were paid for participating in the experiments.
    All volunteers were sitting in an armchair, watching a flat screen monitor
    placed approximately 1 m away at eye level. For each subject 5 sessions
    are provided, whereby the first two sessions contain training data without
    feedback (screening), and the last three sessions were recorded with
    feedback.

    Three bipolar recordings (C3, Cz, and C4) were recorded with a sampling
    frequency of 250 Hz.They were bandpass- filtered between 0.5 Hz and 100 Hz,
    and a notch filter at 50 Hz was enabled.  The placement of the three
    bipolar recordings (large or small distances, more anterior or posterior)
    were slightly different for each subject (for more details see [1]).
    The electrode position Fz served as EEG ground. In addition to the EEG
    channels, the electrooculogram (EOG) was recorded with three monopolar
    electrodes.

    The cue-based screening paradigm consisted of two classes,
    namely the motor imagery (MI) of left hand (class 1) and right hand
    (class 2).
    Each subject participated in two screening sessions without feedback
    recorded on two different days within two weeks.
    Each session consisted of six runs with ten trials each and two classes of
    imagery.  This resulted in 20 trials per run and 120 trials per session.
    Data of 120 repetitions of each MI class were available for each person in
    total.  Prior to the first motor im- agery training the subject executed
    and imagined different movements for each body part and selected the one
    which they could imagine best (e. g., squeezing a ball or pulling a brake).

    Each trial started with a fixation cross and an additional short acoustic
    warning tone (1 kHz, 70 ms).  Some seconds later a visual cue was presented
    for 1.25 seconds.  Afterwards the subjects had to imagine the corresponding
    hand movement over a period of 4 seconds.  Each trial was followed by a
    short break of at least 1.5 seconds.  A randomized time of up to 1 second
    was added to the break to avoid adaptation

    For the three online feedback sessions four runs with smiley feedback
    were recorded, whereby each run consisted of twenty trials for each type of
    motor imagery.  At the beginning of each trial (second 0) the feedback (a
    gray smiley) was centered on the screen.  At second 2, a short warning beep
    (1 kHz, 70 ms) was given. The cue was presented from second 3 to 7.5. At
    second 7.5 the screen went blank and a random interval between 1.0 and 2.0
    seconds was added to the trial.

    References
    ----------
    .. [1] R. Leeb, F. Lee, C. Keinrath, R. Scherer, H. Bischof,
        G. Pfurtscheller. Brain-computer communication: motivation, aim, and
        impact of exploring a virtual apartment. IEEE Transactions on Neural
        Systems and Rehabilitation Engineering 15, 473-482, 2007.

    Parameters
    ----------
    %(subjects)s
    %(epochs_tmin_tmax)s
    %(baseline)s
    %(picks)s
    %(resample)s
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        rename: str | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 10)),
            interval=[3.0, 7.5],
            event_id={"left_hand": 1, "right_hand": 2},
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            baseline=baseline,
            picks=picks,
            resample=resample,
        )
        self._ch_names = ["C3", "Cz", "C4", "EOG1", "EOG2", "EOG3"]
        self._ch_types = ["eeg"] * 3 + ["eog"] * 3
        self._data_url = f"{URL}004-2014/"
        self._data_path = DATA_PATH / "bciciv2b"

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        sessions = []
        for i, r in enumerate(["T", "E"], start=1):
            filename = data_dl(
                f"{self._data_url}B{subject:02d}{r}.mat",
                self._data_path,
                force_update=False,
            )
            runs, _ = _convert_mi(
                filename, self._ch_names, self._ch_types, verbose=verbose
            )
            sessions.extend(runs)

        sessions = {
            f"session_{ii}": {"run_1": run} for ii, run in enumerate(sessions, start=1)
        }
        return sessions



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from sklearn.model_selection import StratifiedKFold
from torchmetrics.aggregation import MeanMetric, CatMetric

from .base import ClsExp
from ..trainer.classifier import BaseClassifier
from ..datasets.base import SplitEEGData
from ..transforms.base import ToEEGData, SplitTrainTest
from ..tools import Filer
from ..utils import DPEEG_SEED, get_init_args
from ..tools.docs import fill_doc


@fill_doc
class KFold(ClsExp):
    r"""K-Fold cross validation experiment.

    The KFold experiment divides the dataset into K non-overlapping subsets
    (i.e., "folds") and repeatedly trains and tests the model. The purpose is
    to reduce the dependence of the model evaluation results on the way the
    dataset is divided and to improve the stability and reliability of the
    evaluation results. However, its computational cost is high, especially for
    large datasets and complex models. It may take a long time to complete the
    training of all folds.

    Two validation methods are provided in the experiment, determined by the
    parameter ``isolate_testset``. When set to ``True``, it indicates that the
    test set is independent of the k-fold cross-validation. That is, for each
    fold, the data is divided into a training set and a validation set to find
    the optimal parameters for each fold, and then the model is evaluated on an
    independent dataset. When set to ``False``, it indicates that one fold of
    data in each fold is used as the test set, and the remaining folds are used
    to train the model. The average value of all folds' evaluations is used as
    the performance metric of the model. The specific experimental method is
    shown in the figure below, which illustrates a 3-fold cross-validation
    experiment:

    .. image:: ../_static/images/kfold_isolate_testset.png
       :align: center
       :alt: kfold isolate testset

    When the training set and test set come from different sessions, setting
    this parameter is very useful.

    Parameters
    ----------
    %(trainer)s
    %(out_folder)s
    k : int, optional
        k of k-Fold.
    isolate_testset : bool
        By default, the test set is independent, that is, the k-fold cross-
        validation at this time only divides the training set and the
        verification set based on the training set to implement an early
        stopping mechanism, and finally evaluates on the isolated test set.
        If False, the test set is for each fold of k-fold cross-validation.
    shuffle : bool
        Shuffle before kfold.
    seed : int
        Seed of random for review.
    %(timestamp)s

    Notes
    -----
    If ``isolate_testset`` False, please provide the ``transforms`` parameter
    of the ``run`` function to avoid data leakage caused by operations such
    as data augmentation in advance. When set to ``True``, it means that the
    experiment requires the `trainer` to support a validation set.
    """

    def __init__(
        self,
        trainer: BaseClassifier,
        out_folder: str | None = None,
        k: int = 5,
        isolate_testset: bool = True,
        shuffle: bool = True,
        seed: int = DPEEG_SEED,
        timestamp: bool = True,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(
            get_init_args(self, locals(), ret_dict=True),
            trainer=trainer,
            out_folder=out_folder,
            timestamp=timestamp,
            verbose=verbose,
        )
        self.k = str(k)
        self.isolate_testset = isolate_testset
        self.skf = StratifiedKFold(k, shuffle=shuffle, random_state=seed)

    def _run_sub_classifier(self, eegdata, sub_folder):
        if self.isolate_testset:
            raise TypeError("Isolated test set is useless with a `Classifier` trainer.")

        eegdata = ToEEGData()(eegdata, verbose=False)

        result = {}
        filer = Filer(sub_folder / "summary.txt")
        train_acc_metric = MeanMetric()
        test_acc_metric = MeanMetric()
        preds_metric = CatMetric()
        target_metric = CatMetric()

        for exp_idx, (train_idx, test_idx) in enumerate(
            self.skf.split(eegdata["edata"], eegdata["label"]), start=1
        ):
            self.logger.info(f"\n# ---- {sub_folder.name}_exp{exp_idx} ---- #")

            trainset = eegdata.index(train_idx)
            testset = eegdata.index(test_idx)
            egd = self._trans_eegdata(SplitEEGData(trainset, testset))
            train_set = egd["train"]
            test_set = egd["test"]

            exp_folder = sub_folder / f"exp{exp_idx}"
            exp_result = self.trainer.fit(
                trainset=train_set,
                testset=test_set,
                log_dir=exp_folder,
            )

            result[f"exp{exp_idx}"] = exp_result
            train_acc_metric.update(exp_result["train"]["acc"])
            test_acc_metric.update(exp_result["test"]["acc"])
            preds_metric.update(exp_result["test"]["preds"])
            target_metric.update(exp_result["test"]["target"])

            filer.write(
                f"Exp_{str(exp_idx).zfill(len(self.k))} Acc: "
                f"Train={exp_result['train']['acc']:.4f} | "
                f"Test={exp_result['test']['acc']:.4f}\n"
            )

        train_acc = train_acc_metric.compute()
        test_acc = test_acc_metric.compute()
        filer.write(f"Avg Acc = {test_acc*100:.2f}%\n")
        self.logger.info("-" * 30)
        self.logger.info(f"Avg Acc: Train={train_acc:.4f} | Test={test_acc:.4f}")

        result.update(
            {
                "acc": test_acc,
                "preds": preds_metric.compute(),
                "target": target_metric.compute(),
            }
        )
        return result

    def _run_sub_classifier_two_stage(self, eegdata, sub_folder):
        if self.isolate_testset:
            eegdata = self._trans_eegdata(eegdata)
            X = eegdata["train"]["edata"]
            y = eegdata["train"]["label"]
        else:
            eegdata = ToEEGData()(eegdata, verbose=False)
            X = eegdata["edata"]
            y = eegdata["label"]

        result = {}
        filer = Filer(sub_folder / "summary.txt")
        train_acc_metric = MeanMetric()
        valid_acc_metric = MeanMetric()
        test_acc_metric = MeanMetric()
        preds_metric = CatMetric()
        target_metric = CatMetric()

        for exp_idx, (train_idx, test_idx) in enumerate(self.skf.split(X, y), start=1):
            self.logger.info(f"\n# ---- {sub_folder.name}_exp{exp_idx} ---- #")

            if self.isolate_testset:
                train_set = eegdata["train"].index(train_idx)
                valid_set = eegdata["train"].index(test_idx)
                test_set = eegdata["test"]
            else:
                trainset = eegdata.index(train_idx)  # type: ignore
                testset = eegdata.index(test_idx)  # type: ignore
                egd = self._trans_eegdata(SplitEEGData(trainset, testset))

                test_set = egd["test"]
                trainset = egd["train"]
                egd = SplitTrainTest()(trainset)
                train_set = egd["train"]
                valid_set = egd["test"]

            exp_folder = sub_folder / f"exp{exp_idx}"
            exp_result = self.trainer.fit(
                trainset=train_set,
                validset=valid_set,
                testset=test_set,
                log_dir=exp_folder,
            )

            result[f"exp{exp_idx}"] = exp_result
            train_acc_metric.update(exp_result["train"]["acc"])
            valid_acc_metric.update(exp_result["valid"]["acc"])
            test_acc_metric.update(exp_result["test"]["acc"])
            preds_metric.update(exp_result["test"]["preds"])
            target_metric.update(exp_result["test"]["target"])

            filer.write(
                f"Exp_{str(exp_idx).zfill(len(self.k))} Acc: "
                f"Train={exp_result['train']['acc']:.4f} | "
                f"Valid={exp_result['train']['acc']:.4f} | "
                f"Test={exp_result['test']['acc']:.4f}\n"
            )

        train_acc = train_acc_metric.compute()
        valid_acc = valid_acc_metric.compute()
        test_acc = test_acc_metric.compute()
        filer.write(f"Avg Acc = {test_acc*100:.2f}%\n")
        self.logger.info("-" * 30)
        self.logger.info(
            f"Avg Acc: Train={train_acc:.4f} | Valid={valid_acc:.4f} | "
            f"Test={test_acc:.4f}"
        )

        result.update(
            {
                "acc": test_acc,
                "preds": preds_metric.compute(),
                "target": target_metric.compute(),
            }
        )
        return result

    def _run_sub(self, eegdata, sub_folder):
        """Basic K-Fold cross validation function.

        Returns
        -------
        Return test_acc, test_kappa, test_preds, test_target and results :\n
        results = {
            'expNo_1' : { ... },
            'expNo_2' : { ... },
            .
            .
        }
        """
        self.timer.start("kfold")
        self.logger.info(f"\n# ---------- {sub_folder.name} ---------- #")
        result = self._run_sub_func(eegdata, sub_folder)
        h, m, s = self.timer.stop("kfold")
        self.logger.info(
            f"\n[{self.k}Fold CV Finish] - [Cost Time = {h}H:{m}M:{s:.2f}S]"
        )

        return result["acc"], result["preds"], result["target"], result



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import json
from pathlib import Path

from tqdm import tqdm
from mne.utils import verbose, logger

from .base import BaseDataset
from .utils import check_inter_and_compl, load
from ..utils import get_init_args
from ..tools.docs import fill_doc


@fill_doc
class LoadDataset(BaseDataset):
    """Read a dataset from a local folder.

    This will read the dataset that has already been saved from the local
    folder. Use with :class:`dpeeg.save_dataset`.

    Parameters
    ----------
    folder : str, Path
        The path to the saved dataset.
    %(subjects)s
    %(rename)s
    """

    @verbose
    def __init__(
        self,
        folder: str | Path,
        subjects: list[int] | None = None,
        rename: str | None = None,
        verbose=None,
    ) -> None:
        super().__init__(get_init_args(self, locals(), rename=rename, ret_dict=True))
        self.folder = Path(folder).resolve(strict=True)

        dataset_info_path = self.folder / "dataset_info.json"
        if dataset_info_path.exists():
            with open(dataset_info_path, "r") as filer:
                dataset_info = json.load(filer)
            if rename is None:
                self._repr["_obj_name"] = dataset_info["name"]
            self.event_id = dataset_info["event_id"]

        else:
            logger.warning("'dataset_info' file not found, using default parameters.")
            self.event_id = None

        path_list = [p for p in self.folder.iterdir() if p.name != "dataset_info.json"]
        subject_list = list([int(p.stem.split("_")[1]) for p in path_list])
        subject_list.sort()

        if subjects is None:
            self.subjects = subject_list
        else:
            self.subjects, _ = check_inter_and_compl(
                subjects, subject_list, verbose=verbose
            )

        self._repr["subjects"] = len(self.subjects)

    def _get_single_subject_data(self, subject: int):
        subject_path = self.folder / f"sub_{subject}.egd"
        return load(subject_path)

    def get_data(self, progressbar: bool = True):
        """Returns the eegdata of all subjects."""
        subjects = tqdm(
            self.subjects, "Load EEGData", unit="sub", disable=not progressbar
        )

        data = {}
        for subject in subjects:
            data[subject] = self._get_single_subject_data(subject)
        return data

    def keys(self):
        return self.subjects



__version__ = "0.4.2"

from .datasets import EEGData, MultiSessEEGData, SplitEEGData, save, load
from . import transforms, exps, models, trainer, tools
from .utils import set_log_level, set_seed

# :wq



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

# This code is adapted from https://github.com/ravikiran-mane/FBCNet/blob/master/codes/centralRepo/stopCriteria.py
# The original code is licensed under the MIT License.

import abc
import os, sys, json


CURRENT_MODULE = sys.modules[__name__]


class Criteria(abc.ABC):
    @abc.abstractmethod
    def __call__(self, variables: dict) -> bool:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    def reset_parameters(self) -> None:
        pass


class ComposeStopCriteria(Criteria):
    """Advanced complex logical expression parser to compose stopping criteria.

    Parameters
    ----------
    stopcri : dict, str
        Type is str, indicate the configuration file path. Type is dict,
        indicate a dictionary containing initialization parameters for the
        corresponding stopping criteria.

    Returns
    -------
    bool :
        Whether the stopping condition is reached.
    """

    def __init__(self, stopcri: dict | str) -> None:
        if isinstance(stopcri, str):
            with open(os.path.abspath(stopcri)) as fp:
                cri: dict = json.load(fp)
        else:
            cri: dict = stopcri
        self.call: Criteria = CURRENT_MODULE.__dict__[list(cri.keys())[0]](
            **cri[list(cri.keys())[0]]
        )

    def __call__(self, variables: dict) -> bool:
        return self.call(variables)

    def __repr__(self) -> str:
        return str(self.call)

    def reset_parameters(self) -> None:
        self.call.reset_parameters()


class And(Criteria):
    """And operation on two stop criteria.

    Parameters
    ----------
    cri1 : dict
        Dictionary describing first criteria.
    cri2 : dict
        Dictionary describing second criteria.

    Notes
    -----
    If you wish to do and on multiple cases then do like: And(And(A, B), C)...
    """

    def __init__(self, cri1, cri2) -> None:
        self.cri1: Criteria = CURRENT_MODULE.__dict__[list(cri1.keys())[0]](
            **cri1[list(cri1.keys())[0]]
        )
        self.cri2: Criteria = CURRENT_MODULE.__dict__[list(cri2.keys())[0]](
            **cri2[list(cri2.keys())[0]]
        )

    def __call__(self, variables: dict) -> bool:
        return self.cri1(variables) and self.cri2(variables)

    def __repr__(self) -> str:
        return f"({self.cri1} 'AND' {self.cri2})"

    def reset_parameters(self) -> None:
        self.cri1.reset_parameters()
        self.cri2.reset_parameters()


class Or(Criteria):
    """Or operation on two stop criteria.

    Parameters
    ----------
    cri1 : dict
        Dictionary describing first criteria.
    cri2 : dict
        Dictionary describing second criteria.

    Notes
    -----
    If you wish to do and on multiple cases then do like:
    And(And(A, B), C)...
    """

    def __init__(self, cri1: dict, cri2: dict) -> None:
        self.cri1: Criteria = CURRENT_MODULE.__dict__[list(cri1.keys())[0]](
            **cri1[list(cri1.keys())[0]]
        )
        self.cri2: Criteria = CURRENT_MODULE.__dict__[list(cri2.keys())[0]](
            **cri2[list(cri2.keys())[0]]
        )

    def __call__(self, variables: dict) -> bool:
        return self.cri1(variables) or self.cri2(variables)

    def __repr__(self) -> str:
        return f"({self.cri1} 'OR' {self.cri2})"

    def reset_parameters(self) -> None:
        self.cri1.reset_parameters()
        self.cri2.reset_parameters()


class MaxEpoch(Criteria):
    """Stop when given number of epochs reached.

    Parameters
    ----------
    max_epochs : int
        Maximum epochs to watch.
    var_name : str
        Key name to compare with in the variables dictionary.
    """

    def __init__(self, max_epochs: int, var_name: str) -> None:
        self.max_epochs = max_epochs
        self.var_name = var_name

    def __call__(self, variables: dict) -> bool:
        return variables[self.var_name] >= self.max_epochs

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(max_epochs={self.max_epochs}"
            f", var_name={self.var_name})"
        )


class NoDecrease(Criteria):
    """Stop if there is no decrease on a given monitor channel for given
    number of epochs.

    Parameters
    ----------
    num_epochs : int
        Number of epochs to wait while there is no decrease in the value.
    var_name : str
        Key name to compare with in the variables dictionary.
    min_change : float
        Minimum relative decrease which resets the num_epochs.
    """

    def __init__(self, num_epochs: int, var_name: str, min_change=1e-6):
        self.num_epochs = num_epochs
        self.var_name = var_name
        self.min_change = min_change
        self.min_value = float("inf")
        self.current_epoch = 0

    def __call__(self, variables: dict) -> bool:
        var = variables[self.var_name]
        devar = (
            (1 - self.min_change) * self.min_value
            if self.min_value > 0
            else (1 + self.min_change) * self.min_value
        )
        if var <= devar and var != 0:
            self.min_value = variables[self.var_name]
            self.current_epoch = 1
        else:
            self.current_epoch += 1

        return self.current_epoch >= self.num_epochs

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(num_epochs={self.num_epochs}"
            f", var_name={self.var_name}, min_change={self.min_change})"
        )

    def reset_parameters(self) -> None:
        self.min_value = float("inf")
        self.current_epoch = 0


class Bigger(Criteria):
    """Stop when greater than the specified value.

    Parameters
    ----------
    var : float
        Maximum value to watch.
    var_name : str
        Key name to compare with in the variables dictionary.
    """

    def __init__(self, var: float, var_name: str) -> None:
        self.var = var
        self.var_name = var_name

    def __call__(self, variables: dict) -> bool:
        return variables[self.var_name] > self.var

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(var={self.var})"


class Smaller(Criteria):
    """Stop when less than the specified value.

    Parameters
    ----------
    var : float
        Minimum value to watch.
    var_name : str
        Key name to compare with in the variables dictionary.
    """

    def __init__(self, var: float, var_name: str) -> None:
        self.var = var
        self.var_name = var_name

    def __call__(self, variables: dict) -> bool:
        return variables[self.var_name] < self.var

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(var={self.var})"



import numpy as np
from mne.io import read_raw_edf
from mne.channels import make_standard_montage

from .base import EpochsDataset, DATA_PATH
from .download import data_dl
from ..utils import get_init_args
from ..tools.docs import fill_doc


URL = "https://physionet.org/files/"


@fill_doc
class PhysioNet_MI(EpochsDataset):
    """Physionet Motor Imagery dataset.

    .. admonition:: Dataset summary

        ====== ====== ======= ===== ======== ======
        Subj   Chan   Time    Cls   Freq     Sess
        ====== ====== ======= ===== ======== ======
        109    64     3.0 s   4     160 Hz   1
        ====== ====== ======= ===== ======== ======

    Physionet MI dataset: https://physionet.org/pn4/eegmmidb/

    This data set consists of over 1500 one- and two-minute EEG recordings,
    obtained from 109 volunteers [2]_.

    Subjects performed different motor/imagery tasks while 64-channel EEG were
    recorded using the BCI2000 system (http://www.bci2000.org) [1]_.
    Each subject performed 14 experimental runs: two one-minute baseline runs
    (one with eyes open, one with eyes closed), and three two-minute runs of
    each of the four following tasks:

    1. A target appears on either the left or the right side of the screen.
    The subject opens and closes the corresponding fist until the target
    disappears. Then the subject relaxes.

    2. A target appears on either the left or the right side of the screen.
    The subject imagines opening and closing the corresponding fist until
    the target disappears. Then the subject relaxes.

    3. A target appears on either the top or the bottom of the screen.
    The subject opens and closes either both fists (if the target is on top)
    or both feet (if the target is on the bottom) until the target
    disappears. Then the subject relaxes.

    4. A target appears on either the top or the bottom of the screen.
    The subject imagines opening and closing either both fists
    (if the target is on top) or both feet (if the target is on the bottom)
    until the target disappears. Then the subject relaxes.

    references
    ----------
    .. [1] Schalk, G., McFarland, D.J., Hinterberger, T., Birbaumer, N. and
        Wolpaw, J.R., 2004. BCI2000: a general-purpose brain-computer
        interface (BCI) system. IEEE Transactions on biomedical engineering,
        51(6), pp.1034-1043.
    .. [2] Goldberger, A.L., Amaral, L.A., Glass, L., Hausdorff, J.M., Ivanov,
        P.C., Mark, R.G., Mietus, J.E., Moody, G.B., Peng, C.K., Stanley,
        H.E. and PhysioBank, P., PhysioNet: components of a new research
        resource for complex physiologic signals Circulation 2000 Volume
        101 Issue 23 pp. E215-E220.

    Parameters
    ----------
    %(subjects)s
    %(epochs_tmin_tmax)s
    %(baseline)s
    %(picks)s
    %(resample)s
    imagined: bool (default True)
        if True, return runs corresponding to motor imagination.
    executed: bool (default False)
        if True, return runs corresponding to motor execution.
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        imagined: bool = True,
        executed: bool = False,
        rename: str | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 110)),
            interval=[0.0, 3.0],
            event_id={
                "rest": 1,
                "left_hand": 2,
                "right_hand": 3,
                "hands": 4,
                "feet": 5,
            },
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            baseline=baseline,
            picks=picks,
            resample=resample,
        )
        self._data_path = DATA_PATH / "physionet_mi"
        self._data_url = f"{URL}eegmmidb/1.0.0/"

        # fmt: off
        self._renames = {
            "AFZ": "AFz", "PZ": "Pz", "FPZ": "Fpz", "FCZ": "FCz", "FP1": "Fp1", 
            "CZ": "Cz", "OZ": "Oz", "POZ": "POz", "IZ": "Iz", "CPZ": "CPz", 
            "FP2": "Fp2", "FZ": "Fz",
        }
        # fmt: on
        self._montage = make_standard_montage("standard_1005")

        self._feet_runs = []
        self._hand_runs = []
        if imagined:
            self._feet_runs += [6, 10, 14]
            self._hand_runs += [4, 8, 12]
        if executed:
            self._feet_runs += [5, 9, 13]
            self._hand_runs += [3, 7, 11]

    def _load_run(self, subject, run, verbose):
        filename = data_dl(
            f"{self._data_url}S{subject:03d}/S{subject:03d}R{run:02d}.edf",
            self._data_path,
            force_update=False,
        )
        raw = read_raw_edf(filename, preload=True, verbose=verbose)
        raw.rename_channels(lambda x: x.strip("."))
        raw.rename_channels(lambda x: x.upper())
        raw.rename_channels(self._renames)
        raw.set_montage(self._montage)
        return raw

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        session = {}

        # hand runs
        idx = 1
        for run in self._hand_runs:
            raw = self._load_run(subject, run, verbose)
            stim = raw.annotations.description.astype(np.dtype("<U10"))
            stim[stim == "T0"] = "rest"
            stim[stim == "T1"] = "left_hand"
            stim[stim == "T2"] = "right_hand"
            raw.annotations.description = stim
            session[f"run_{idx}"] = raw
            idx += 1

        # feet runs
        for run in self._feet_runs:
            raw = self._load_run(subject, run, verbose)
            stim = raw.annotations.description.astype(np.dtype("<U10"))
            stim[stim == "T0"] = "rest"
            stim[stim == "T1"] = "hands"
            stim[stim == "T2"] = "feet"
            raw.annotations.description = stim
            session[f"run_{idx}"] = raw
            idx += 1

        return {"session_1": session}



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..tools.docs import fill_doc


__all__ = ["LightConvNet"]


class LogVarLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        out = torch.clamp(x.var(dim=self.dim), 1e-6, 1e6)
        return torch.log(out)


class LightweightConv1d(nn.Module):
    """
    Args:
        input_size: # of channels of the input and output
        kernel_size: convolution channels
        padding: padding
        num_heads: number of heads used. The weight is of shape
            `(num_heads, 1, kernel_size)`
        weight_softmax: normalize the weight with softmax before the convolution
    Shape:
        Input: BxCxT, i.e. (batch_size, input_size, timesteps)
        Output: BxCxT, i.e. (batch_size, input_size, timesteps)
    Attributes:
        weight: the learnable weights of the module of shape
            `(num_heads, 1, kernel_size)`
        bias: the learnable bias of the module of shape `(input_size)`
    """

    def __init__(
        self,
        input_size,
        kernel_size=1,
        padding=0,
        heads=1,
        weight_softmax=False,
        bias=False,
    ):
        super().__init__()
        self.input_size = input_size
        self.kernel_size = kernel_size
        self.heads = heads
        self.padding = padding
        self.weight_softmax = weight_softmax
        self.weight = nn.Parameter(torch.Tensor(heads, 1, kernel_size))

        if bias:
            self.bias = nn.Parameter(torch.Tensor(input_size))
        else:
            self.bias = None

        self.init_parameters()

    def init_parameters(self):
        nn.init.xavier_uniform_(self.weight)
        if self.bias is not None:
            nn.init.constant_(self.bias, 0.0)

    def forward(self, input):
        B, C, T = input.size()
        H = self.heads

        weight = self.weight
        if self.weight_softmax:
            weight = F.softmax(weight, dim=-1)

        input = input.view(-1, H, T)
        output = F.conv1d(input, weight, padding=self.padding, groups=self.heads)
        output = output.view(B, C, -1)
        if self.bias is not None:
            output = output + self.bias.view(1, -1, 1)

        return output


@fill_doc
class LightConvNet(nn.Module):
    """A Temporal Dependency Learning CNN With Attention Mechanism for MI-EEG
    Decoding (LightConvNet).

    LightConvNet [1]_ first implements the spatial convolution to learn spatial
    and spectral information from multi-view EEG data, which is preprocessed
    with a filter bank. Then, LightConvNet employs a series of non-overlapped
    time windows to segment the output time series. The discriminative feature
    from each time window is further extracted using a temporal variance layer
    to capture MI-related patterns in different stages during MI tasks.
    Moreover, LightConvNet designs a novel temporal attention module to further
    learn temporal dependencies among discriminative features from different
    time windows. The temporal attention module assigns different weights to
    features in various time windows according to their contribution to the
    final decoding performance, and fuses them into more discriminative
    features. Finally, the fused features are used for classification.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    bands : int
        The filter dimension of the input multi-view data.
    embed_dim : int
        Number of spatial filters.
    win_len : int
        The length of the time window.
    heads : int
        Number of multi-head attention.
    weight_softmax : bool
        Normalize the weight with softmax before the convolution.
    bias : bool
        The learnable bias.

    References
    ----------
    .. [1] “A Temporal Dependency Learning CNN With Attention Mechanism for
        MI-EEG Decoding | IEEE Journals & Magazine | IEEE Xplore.” Accessed:
        Oct. 20, 2023.
        [Online]. Available: https://ieeexplore.ieee.org/document/10196350
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        bands: int = 9,
        embed_dim: int = 64,
        win_len: int = 250,
        heads: int = 8,
        weight_softmax: bool = True,
        bias: bool = False,
    ):
        super().__init__()
        self.win_len = win_len

        self.spacial_block = nn.Sequential(
            nn.Conv2d(bands, embed_dim, (nCh, 1)), nn.BatchNorm2d(embed_dim), nn.ELU()
        )

        self.temporal_block = LogVarLayer(dim=3)

        self.conv = LightweightConv1d(
            embed_dim,
            (nTime // win_len),
            heads=heads,
            weight_softmax=weight_softmax,
            bias=bias,
        )

        self.classify = nn.Sequential(nn.Linear(embed_dim, nCls), nn.LogSoftmax(dim=1))

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, bands, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        out = self.spacial_block(x)
        out = out.reshape([*out.shape[0:2], -1, self.win_len])
        out = self.temporal_block(out)
        out = self.conv(out)
        out = out.view(out.size(0), -1)
        out = self.classify(out)
        return out



from .EEGNet import EEGNet
from .EEGConformer import EEGConformer
from .FBCNet import FBCNet
from .IFNet import IFNet, IFNetAdamW
from .LightConvNet import LightConvNet
from .MSVTNet import MSVTNet, JointCrossEntoryLoss
from .ShallowConvNet import ShallowConvNet
from .DeepConvNet import DeepConvNet



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import time
from datetime import datetime


class Timer:
    """Timing the running time of the program.

    Examples
    --------
    If you want to start a default timer and then calculate the program interval,
    you can do like:

    >>> timer = Timer()
    >>> do_something()
    >>> h, m, s = timer.stop()
    >>> h, m, s
    1, 2, 13.435

    And you can update the start time of the default timer, just do like:

    >>> timer.start()

    You can also register a specified timer and update the start time of it.

    >>> timer.start('new')
    >>> h, m, s = timer.stop('new')
    """

    def __init__(self, name: str = "dpeeg_timer") -> None:
        """Timing the running time of the program.
        Initialization will automatically start a default timer (dpeeg_timer).
        """
        self._beg = {}
        self.start(name)
        self.start("dpeeg_timer")

    def start(self, name: str = "dpeeg_timer") -> str:
        """Update specified timer or register a timer if it does not exist."""
        self._beg[name] = time.time()
        return self.ctime()

    def stop(
        self, name: str = "dpeeg_timer", restart: bool = False
    ) -> tuple[int, int, float]:
        """Get the hours, minutes, and seconds of a specific timer."""
        sec = self.stop_precise(name, restart)
        m, s = divmod(sec, 60)
        h, m = divmod(m, 60)

        return int(h), int(m), s

    def stop_precise(self, name: str = "dpeeg_timer", restart: bool = False) -> float:
        """Get the precise time for a specific timer."""
        if name not in self._beg.keys():
            raise RuntimeError(f"Timer `{name}` has not been registered yet.")
        ct = time.time()

        if restart:
            self.start(name)

        return ct - self._beg[name]

    @staticmethod
    def ctime() -> str:
        """Return the current local time as a human-readable string."""
        return time.ctime()

    @staticmethod
    def cdate() -> str:
        """Get the current date and time in the 'year-month-day_hour:minute'
        format.
        """
        return datetime.now().strftime("%y-%m-%d_%H:%M")



from setuptools import setup, find_packages

from dpeeg import __version__

install_requires = [
    "numpy>=1.21.5",
    "einops>=0.7.0",
    "seaborn>=0.12.1",
    "pooch>=1.6.0",
    "mne>=1.6",
    "scipy>=1.11.1",
    "scikit-learn>=1.0.2",
    "tqdm>=4.64.1",
    "torchmetrics>=1.0.0",
    "torchinfo>=1.5.0",
    "tensorboard",
    "pandas>=1.5.2",
    "openpyxl",
]

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="dpeeg",
    version=__version__,
    author="SheepTAO",
    author_email="sheeptao@outlook.com",
    license="MIT",
    description="Deep learning with EEG",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SheepTAO/dpeeg",
    packages=find_packages(),
    keywords=["eeg", "deep learning", "pytorch"],
    python_requires=">=3.10",
    install_requires=install_requires,
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Operating System :: OS Independent",
    ],
)



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import sys
import logging
from pathlib import Path


class Logger:
    """Logging hooks for terminals and file streams.

    Can handle terminal printing and file writing in three situations:
    1.Terminal print;
    2.Terminal and file streams write simultaneously;
    3.Ordinary file write.

    Parameters
    ----------
    logger : str
        Specify the logger name, creating it if necessary. If no name is
        specified, return the root logger.
    path : str, optional
        The path of log file.
    mode : str
        The write mode of the log file.
    clevel : int, str
        The log level of console.
    flevel : int, str, optional
        The log level of filehandler. If it is None, the log file will not be
        written and parameter path will be ignored.
    """

    def __init__(
        self,
        logger: str = "dpeeg_root",
        path: str | Path | None = None,
        mode: str = "a",
        clevel: int | str = logging.INFO,
        flevel: int | str | None = None,
    ) -> None:
        self._logger = logging.getLogger(logger)
        self._logger.propagate = False
        self._logger.setLevel(logging.DEBUG)

        self._sh = logging.StreamHandler()
        shfmt = logging.Formatter("%(message)s")
        self._sh.setFormatter(shfmt)
        self._sh.setLevel(clevel)
        self._logger.addHandler(self._sh)
        if flevel is not None:
            assert path is not None, "path cannot be empty when use filer."
            path = Path(path).absolute()
            path.parent.mkdir(parents=True, exist_ok=True)
            self._fh = logging.FileHandler(str(path), mode)
            fhfmt = logging.Formatter(
                "[%(asctime)s] [%(levelname)8s]: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            self._fh.setFormatter(fhfmt)
            self._fh.setLevel(flevel)
            self._logger.addHandler(self._fh)

        sys.excepthook = self.handle_exception

    def update_sh_level(self, level: int | str):
        self._sh.setLevel(level)

    def get_sh_level(self) -> str:
        return logging.getLevelName(self._sh.level)

    def debug(self, message: str):
        self._logger.debug(f"[DEBUG]: {message}")

    def info(self, message: str):
        self._logger.info(message)

    def warning(self, message: str):
        self._logger.warning(f"[WARNING]: {message}")

    def error(self, message: str):
        self._logger.error(f"[ERROR]: {message}")

    def critical(self, message: str):
        self._logger.critical(f"[CRITICAL]: {message}")

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
        self._logger.critical(
            "Uncaught exception>>>", exc_info=(exc_type, exc_value, exc_traceback)
        )



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from typing import Literal

import torch
import numpy as np
import scipy.signal as signal
from numpy import ndarray
from mne.utils import verbose, logger
from torch import Tensor

from ..utils import DPEEG_SEED, iterable_to_str


@verbose
def split_train_test(
    *arrs,
    test_size: float = 0.25,
    seed: int = DPEEG_SEED,
    sample: list[int] | None = None,
    verbose=None,
) -> list:
    """Split an dataset into training and testing sets. The axis along which
    to split is 0.

    .. deprecated:: 0.4.0
        Will be removed in a later version.

    Parameters
    ----------
    *arrs : sequence of indexables with same length / shape[0]
        Allowed inputs are lists and numpy arrays.
    test_size : float
        The proportion of the test set. If index is not None, test_size will be
        ignored. Default use stratified fashion and the last arr serves as the
        class label.
    seed : int
        Random seed when splitting.
    sample : list of int, None
        A list of integers, the entries indicate which data were selected
        as the test set. If None, test_size will be used.

    Returns
    -------
    splitting : list, length=2 * len(arrays)
        List containing train-test split of inputs.
    """
    num_arrs = len(arrs)
    if num_arrs == 0:
        raise ValueError("At least one array required as input.")

    arr_list = [np.array(arr) for arr in arrs]
    lengths = [len(arr) for arr in arr_list]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError(
            "Found input variables with inconsistent numbers of samples: %r"
            % [int(l) for l in lengths]
        )

    if sample is not None:
        smparr = np.array(sample)
        if np.unique(smparr).size != smparr.size:
            raise IndexError(f"Found repeated sampling of test set: {smparr.tolist()}.")
        length = lengths[0]
        if smparr.size >= uniques:
            raise IndexError(
                f"The number of samples (={len(smparr)}) in the "
                "test set cannot exceed the total number of data "
                f"sets (={length})."
            )

        testSample, trainSample = smparr, np.setdiff1d(np.arange(length), smparr)
        res = []
        for arr in arr_list:
            res.extend([arr[trainSample], arr[testSample]])
        return res
    else:
        from sklearn.model_selection import train_test_split

        return train_test_split(
            *arr_list, test_size=test_size, random_state=seed, stratify=arr_list[-1]
        )


@verbose
def merge_train_test(*arrs, verbose=None) -> tuple[ndarray, ndarray]:
    """Merge the data and label of the training set and test set.

    .. deprecated:: 0.4.0
        Will be removed in a later version.

    Parameters
    ----------
    *arrs : sequence of data (ndarray (N, ...)) and label (ndarray (N,))
        Sequence consisting of each piece of data and its corresponding label.

    Returns
    -------
    data, label : ndarray, ndarray
        Merged data and label.

    Examples
    --------
    >>> data = np.arange(6).reshape(2, 3)
    >>> label = np.arange(2)
    >>> train, test = [data, label], [data, label]
    >>> merge_train_test(train, test)
    (array([[3, 4, 5],
            [0, 1, 2],
            [0, 1, 2],
            [3, 4, 5]]),
    array([1, 0, 0, 1]))
    """
    if len(arrs) == 0:
        raise ValueError("At least one array required as input.")

    merge_data = []
    merge_label = []
    for i, arr in enumerate(arrs):
        if not (isinstance(arr, (list, tuple)) and len(arr) == 2):
            raise ValueError(f"Missing data or label in {i}^th group.")

        data, label = arr[0], arr[1]
        if data.shape[0] != label.shape[0]:
            logger.warning(f"Length of data and label in {i}^th group is not uniform.")
        merge_data.append(data)
        merge_label.append(label)

    merge_data = np.concatenate(merge_data)
    merge_label = np.concatenate(merge_label)
    shuffle = np.random.permutation(merge_label.shape[0])
    merge_data = merge_data[shuffle]
    merge_label = merge_label[shuffle]

    return merge_data, merge_label


@verbose
def to_tensor(
    data: Tensor | ndarray, label: Tensor | ndarray, verbose=None
) -> tuple[Tensor, Tensor]:
    """Convert the numpy data and label into trainable Tensor format."""
    # https://discuss.pytorch.org/t/torch-from-numpy-not-support-negative-strides/3663
    dataT = (
        torch.from_numpy(np.ascontiguousarray(data)).float()
        if isinstance(data, ndarray)
        else data.float()
    )
    labelT = (
        torch.from_numpy(np.ascontiguousarray(label)).long()
        if isinstance(label, ndarray)
        else label.long()
    )
    return dataT, labelT


@verbose
def crop(
    data: ndarray,
    tmin: int = 0,
    tmax: int | None = None,
    include_tmax: bool = False,
    verbose=None,
) -> ndarray:
    """Crop a time interval from the data along the last dimension by default.

    Parameters
    ----------
    tmin : int
        Start time of selection in sampling points.
    tmax : int, None
        End time of selection in sampling points. None means use the full time.
    include_tmax : bool
        If `False`, exclude tmax.

    Returns
    -------
    data : ndarray
        The cropped data.
    """
    if tmax is not None and include_tmax:
        tmax += 1
        assert tmax > tmin, "tmax must be greater than tmin"
    return data[..., tmin:tmax]


@verbose
def z_score_norm(
    data: ndarray,
    mean: ndarray | None = None,
    std: ndarray | None = None,
    dim: int | None = None,
    ret: bool = False,
    verbose=None,
):
    r"""Z-score normalization.

    .. math::

       \mathbf{z} = \frac{\mathbf{x} - \mu}{\sqrt{\sigma^{2}}}

    where :math:`\mathbf{x}` and :math:`\mathbf{z}` denote the input data and
    the output of normalization, respectively. :math:`\mu` and :math:`\sigma^2`
    represent the mean and variance values of the sample.

    Parameters
    ----------
    data : ndarray
        Input data.
    mean : ndarray, optional
        The mean used in the normalization process. If None, use the statistics
        of the current sample for normalization.
    std : ndarray, optional
        The standard deviation used in the normalization process. If None, use
        the statistics of the current sample for normalization.
    dim : int, optional
        The dimension to normalize. If None, the entire data is normalized.
    ret : bool, optional
        Returns the corresponding calculated parameter.

    Returns
    -------
    ndarray
        Normalized data. Mean and std of data if ``ret`` is True.
    """
    keepdims = False if dim is None else True
    mean = np.mean(data, dim, keepdims=keepdims) if mean is None else mean
    std = np.std(data, dim, keepdims=keepdims) if std is None else std
    data = (data - mean) / std

    if ret:
        return data, mean, std
    else:
        return data


@verbose
def min_max_norm(
    data: ndarray,
    min: ndarray | None = None,
    max: ndarray | None = None,
    dim: int | None = None,
    ret: bool = False,
    verbose=None,
):
    r"""Min-max normalization.

    .. math::

       \mathbf{z} =
       \frac{\mathbf{x}-\mathbf{x}_{min}}{\mathbf{x}_{max}-\mathbf{x}_{min}}

    where :math:`\mathbf{x}` and :math:`\mathbf{z}` denote the input data and
    the output of normalization, respectively. :math:`\mathbf{x}_{max}` and
    :math:`\mathbf{x}_{min}` represent the maximum and minimum values of the
    sample.

    Parameters
    ----------
    data : ndarray
        Input data.
    min : ndarray, optional
        The minimum used in the normalization process. If None, use the
        statistics of the current sample for normalization.
    max : ndarray, optional
        The maximum used in the normalization process. If None, use the
        statistics of the current smaple for normalization.
    dim : int, None
        The dimension to normalize. If None, the entire data is normalized.

    Returns
    -------
    ndarray
        Normalized data. Minimum and maximum of data if ``ret`` is True.
    """
    keepdims = False if dim is None else True
    minimum = np.min(data, dim, keepdims=keepdims) if min is None else min
    maximum = np.max(data, dim, keepdims=keepdims) if max is None else max
    data = (data - minimum) / (maximum - minimum)

    if ret:
        return data, minimum, maximum
    else:
        return data


@verbose
def slide_win(
    data: ndarray,
    win: int,
    overlap: int = 0,
    label: ndarray | None = None,
    verbose=None,
) -> tuple[ndarray, ndarray | None]:
    """This transform is only splits the time series (dim = -1) through the
    sliding window operation on the original dataset. If the time axis is not
    divisible by the sliding window, the last remaining time data will be
    discarded.

    Parameters
    ----------
    data : ndarray (N, ..., T)
        The data to split. Shape as `(N, ..., T)`, with `N` the number of data
        and `T` the number of samples.
    win : int
        The size of the sliding window.
    overlap : int
        The amount of overlap between adjacent sliding windows.
    label : ndarray (N,), None
        The label of the data. If not None, label will update with sliding window.

    Returns
    -------
    data : ndarray
        Data after sliding window.
    label : ndarray, None
        If label is None, return None. The label corresponding to each window.
    """
    if win < 0 or overlap < 0:
        raise ValueError(
            "win and overlap only allow positive numbers, "
            f"but got {win} and {overlap} respectively."
        )

    if overlap >= win:
        raise ValueError(f"overlap={overlap} should be less than win={win}.")

    if isinstance(label, ndarray) and len(data) != len(label):
        raise ValueError("The number of label and data must be the same.")

    end = win
    times = data.shape[-1]
    if end > times:
        logger.warning("The window is larger than the times to be split.")
        return data, label

    sld_num = 0
    data_list = []
    while end <= times:
        data_list.append(data[..., end - win : end])
        sld_num += 1
        end += win - overlap

    data = np.concatenate(data_list)
    shuffle = np.random.permutation(len(data))
    data = data[shuffle]
    if isinstance(label, ndarray):
        label = np.repeat(label, sld_num)
        label = label[shuffle]
    return data, label


@verbose
def segmentation_and_reconstruction_time(
    data: ndarray,
    label: ndarray,
    samples: int = 125,
    multiply: float = 1.0,
    seed: int = DPEEG_SEED,
    verbose=None,
) -> tuple[ndarray, ndarray]:
    """Signal Segmentation and Recombination in Time Domain.

    This approach is to first divide each EEG trial into several segments and
    then generate new artificial trials as a concatenation of segments coming
    from different and randomly selected trials from the same class while
    maintaining the original time order [1]_.

    Parameters
    ----------
    data : ndarray (N, ..., T)
        Data that will be segmented and randomly recombined. Shape as `(N, ...,
        T)`, with `N` the number of data and `T` the number of samples.
    label : ndarray (N,)
        The label corresponding to the data. Shape as `(N)`.
    samples : int
        The number of consecutive samples to segment the data. eg, 125 for
        250Hz data is segmented by 0.5s.
    multiply : float
        Data expansion multiple of relative metadata, 1 means doubled.
    seed : int
        Controls the shuffling applied to the data after picking.

    Notes
    -----
    The function does not judge the multiples or the number of segmentation.

    References
    ----------
    .. [1] F. Lotte, “Signal Processing Approaches to Minimize or Suppress
        Calibration Time in Oscillatory Activity-Based Brain-Computer
        Interfaces,” Proc. IEEE, vol. 103, no. 6, pp. 871-890, Jun. 2015,
        doi: 10.1109/JPROC.2015.2404941.
    """
    assert samples >= 1, "samples should be at least 1"
    assert multiply > 0, "multiply should be greater than 0"

    parts = data.shape[-1] // samples
    aug_data, aug_label = [], []

    for lb in np.unique(label):
        idx = np.where(label == lb)
        org_data, org_label = data[idx], label[idx]
        aug_num = int(org_data.shape[0] * multiply)
        tmp_aug_data = np.empty((aug_num, *data.shape[1:]))

        for i in range(aug_num):
            for j in range(parts):
                randIdx = np.random.randint(0, org_data.shape[0], parts)
                tmp_aug_data[i, ..., j * samples : (j + 1) * samples] = org_data[
                    randIdx[j], ..., j * samples : (j + 1) * samples
                ]
            if data.shape[-1] % samples:
                randIdx = np.random.randint(0, org_data.shape[0])
                tmp_aug_data[i, ..., (j + 1) * samples :] = org_data[
                    randIdx, ..., (j + 1) * samples :
                ]

        aug_data.append(tmp_aug_data)
        aug_label.append(np.repeat(lb, aug_num))
        aug_data.append(org_data)
        aug_label.append(org_label)

    aug_data = np.concatenate(aug_data)
    aug_label = np.concatenate(aug_label)

    rng = np.random.RandomState(seed)
    shuffle_idx = rng.permutation(aug_label.shape[0])
    aug_data = aug_data[shuffle_idx]
    aug_label = aug_label[shuffle_idx]

    return aug_data, aug_label


@verbose
def cheby2_filter(
    data: ndarray,
    freq: float,
    l_freq: float | None = None,
    h_freq: float | None = None,
    transition_bandwidth: float = 2.0,
    gstop: float = 30,
    gpass: float = 3,
    filter_type: Literal["filter", "filtfilt"] = "filter",
    verbose=None,
):
    """Filter a signal using cheby2 iir filtering.

    Parameters
    ----------
    data : ndarray
        Filtering is performed in the last dimension (time dimension).
    freq : float
        Data sampling frequency.
    l_freq, h_freq : float, None
        Low and high cut off frequency in hertz. If l_freq is None, the data
        are only low-passed. If h_freq is None,  the data are only high-passed.
        Both parameters cannot be None at the same time.
    transition_bandwidth : float
        The bandwidth (in hertz) of the transition region of the frequency res-
        ponse from the passband to the stopband.
    gstop : float
        The minimum attenuation in the stopband (dB).
    gpass : float
        The maximum loss in the passband (dB).
    filter_type : str
        Filter type, available options are 'filtfilt' and 'filter'.

    Returns
    -------
    data : ndarray
        Data after applying bandpass filter.
    """

    n_freq = freq / 2  # Nyquist frequency

    if (l_freq == 0 or l_freq == None) and (h_freq == None or h_freq >= n_freq):
        logger.warning("Not doing any filtering. Invalid cut-off freq.")
        return data

    # low-passed filter
    elif l_freq == 0 or l_freq == None:
        logger.info(f"{h_freq} Hz lowpass filter.")
        fpass = h_freq / n_freq  # type:ignore
        fstop = (h_freq + transition_bandwidth) / n_freq  # type:ignore
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a = signal.cheby2(N, gstop, fstop, "lowpass")

    # high-passed filter
    elif h_freq == None or h_freq == n_freq:
        logger.info(f"{l_freq} Hz highpass filter.")
        fpass = l_freq / n_freq
        fstop = (l_freq - transition_bandwidth) / n_freq
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a = signal.cheby2(N, gstop, fstop, "highpass")

    # band-passed filter
    else:
        logger.info(f"{l_freq} - {h_freq} Hz bandpass filter.")
        fpass = [l_freq / n_freq, h_freq / n_freq]
        fstop = [
            (l_freq - transition_bandwidth) / n_freq,
            (h_freq + transition_bandwidth) / n_freq,
        ]
        N, Ws = signal.cheb2ord(fpass, fstop, gpass, gstop)
        b, a = signal.cheby2(N, gstop, fstop, "bandpass")

    if filter_type == "filtfilt":
        out = signal.filtfilt(b, a, data)
    else:
        out = signal.lfilter(b, a, data)

    return out


@verbose
def label_mapping(
    label: ndarray,
    mapping: ndarray | None = None,
    order: bool = True,
    verbose=None,
) -> ndarray:
    """Update the original label according to mapping rules.

    Parameters
    ----------
    label : ndarray (N,)
        Original label list.
    mapping : ndarray (2, label_num)
        Label mapping relationship. The first row is the original label, and
        the second row is the mapped label. If ``None``, the label will be
        updated in ascending order starting from 0.
    order : bool
        Force the new labels to start incrementing from 0.

    Returns
    -------
    Returns the mapped label.

    Examples
    --------
    Merge label:

    >>> label = np.array([1, 2, 3, 2, 3, 1, 3, 4])
    >>> mapping = np.array([[1, 2, 3, 4], [0, 1, 0, 1]])
    >>> label_mapping(label, mapping)
    array([0, 1, 0, 1, 0, 0, 0, 1])

    Rearrange the original label:

    >>> mapping = np.array([[1, 2, 3, 4], [3, 2, 1, 0]])
    >>> label_mapping(label, mapping)
    array([3, 2, 1, 2, 1, 3, 1, 0])
    """
    label, uni_label = np.array(label), np.unique(label)

    if mapping is None:
        mapping = np.array([uni_label, np.arange(len(uni_label))])
        order = False
    else:
        mapping, uni_mapping = np.array(mapping), np.unique(mapping[0])

        if mapping.ndim != 2 or mapping.shape[0] != 2:
            raise ValueError("The mapping is not 2D.")

        if len(uni_label) != len(uni_mapping) or any(uni_label != uni_mapping):
            raise ValueError("Mapping does not correspond to label.")

    if order:
        uni_new_label = np.unique(mapping[1])
        order_label = np.arange(len(uni_new_label))
        if any(order_label != uni_new_label):
            raise ValueError("Mapping error, set `order = false` to turn off.")

    new_label = np.empty_like(label)
    for i in range(mapping.shape[1]):
        indices = np.where(mapping[0][i] == label)
        new_label[indices] = mapping[1][i]

    return new_label


@verbose
def pick_label(
    *data: ndarray,
    label: ndarray,
    pick: ndarray,
    order: bool = False,
    shuffle: bool = True,
    seed: int = DPEEG_SEED,
    verbose=None,
):
    """Pick a subset of data by label.

    Pick the required labels and data from the dataset or re-label them.

    Parameters
    ----------
    *data : sequence of indexables with same length / shape[0]
        The data to pick. Shape as `(N, ...)`.
    label : ndarray (N,)
        Dataset label.
    pick : ndarray (n,)
        Label to include.
    order : bool
        If `True`, relabel the selected labels.
    shuffle : bool
        Whether or not to shuffle the data after picking.
    seed : int
        Controls the shuffling applied to the data after picking.

    Returns
    -------
    data, label : list of ndarray, length = len(data) + 1
        Returns the picked data and label.

    Examples
    --------
    >>> data1 = np.arange(24).reshape(8, 3)
    >>> data2 = np.arange(2, 26).reshape(8, 3)
    >>> label = np.array([0, 0, 1, 1, 2, 2, 2, 2])
    >>> pick_label(data1, data2, label=label, pick=np.array([1]))
    ([array([[ 9, 10, 11],
             [ 6,  7,  8]]),
      array([[11, 12, 13],
             [ 8,  9, 10]])],
     array([1, 1]))

    >>> pick_label(data1, label=label, pick=np.array([0, 2]), order=True)
    ([array([[ 0,  1,  2],
             [ 3,  4,  5],
             [21, 22, 23],
             [12, 13, 14],
             [18, 19, 20],
             [15, 16, 17]])],
     array([0, 0, 1, 1, 1, 1]))
    """
    num_data = len(data)
    if num_data == 0:
        raise ValueError("At least one data required as input.")

    elements = set(pick) - set(label)
    if elements:
        raise ValueError(f"{iterable_to_str(elements)} not in label.")

    new_label = []
    new_data = [[] for _ in range(num_data)]
    for i, p in enumerate(np.unique(pick)):
        indices = np.where(label == p)[0]
        for n in range(num_data):
            new_data[n].append(data[n][indices])

        if order:
            new_label.append(np.repeat(i, len(indices)))
        else:
            new_label.append(label[indices])

    new_data = [np.concatenate(nd) for nd in new_data]
    new_label = np.concatenate(new_label)

    if shuffle:
        rng = np.random.RandomState(seed)
        shuffle_idx = rng.permutation(new_label.shape[0])
        new_data = [nd[shuffle_idx] for nd in new_data]
        new_label = new_label[shuffle_idx]

    return new_data, new_label


def pick_data(data: ndarray, labels: ndarray, label: int) -> ndarray:
    """Index data based on specified label.

    Parameters
    ----------
    data : ndarray (N, ...)
        Input data.
    labels : ndarray (N,)
        Label table corresponding to the data.
    label : int
        The label to get.

    Returns
    -------
    Tensor
        All data for the specified label.
    """
    assert label in labels, f"label {label} is not in the labels."

    indices = np.where(labels == label)
    return data[indices]


def _NPMConv(a: ndarray, v: ndarray, mode: Literal["full", "same", "valid"]) -> ndarray:
    """Linear convolution of two multi-dimensional sequences."""
    dim, seq = a.shape[:-1], a.shape[-1]
    a_flatten = a.reshape(-1, seq)
    for i in range(len(a_flatten)):
        a_flatten[i] = np.convolve(a_flatten[i], v, mode)
    return a_flatten.reshape(*dim, seq)


def smooth(signal: ndarray, win: int = 5) -> ndarray:
    """Matlab smooth implement for moving average method in python.

    Parameters
    ----------
    signal : ndarray (N,)
        Input signal.
    win : int
        Dimension of the smoothing window. If you specify win as an even number,
        win is automatically reduced by 1.
    """
    sig_len = len(signal)
    win = win if win % 2 else win - 1
    assert win < sig_len, f"window:{win} exceeds signal length:{sig_len}."

    sig_mid = np.convolve(signal, np.ones(win) / win, mode="valid")
    r = np.arange(1, win - 1, 2)
    sig_beg = np.cumsum(signal[: win - 1])[::2] / r
    sig_end = (np.cumsum(signal[:-win:-1])[::2] / r)[::-1]
    return np.concatenate((sig_beg, sig_mid, sig_end))


def erds_time(
    data: ndarray,
    ref_start: int,
    ref_end: int,
    smooth_win: int,
) -> tuple[ndarray, ndarray]:
    """Time course of ERD/ERS.

    Returns
    -------
    erds : ndarray
        ERD/ERS in %.
    avg_power : ndarray
        Averaged power.
    """
    assert ref_start >= 0, "ref_start is less than 0."
    assert ref_end < data.shape[-1], "ref_end exceeds data time range."
    assert ref_end > ref_start, "ref_end is less than ref_start."

    avg_power = smooth(data.var(axis=0, ddof=1), smooth_win)
    ref_avg_power = np.mean(avg_power[ref_start:ref_end])
    erds = ((avg_power - ref_avg_power) / ref_avg_power) * 100

    return erds, avg_power


def np_flatten(data: np.ndarray, start_dim: int = 0, end_dim: int | None = -1):
    """NumPy implementation of ``torch.flatten``."""
    size = data.shape
    s_size = size[:start_dim]
    e_size = size[end_dim:] if end_dim else ()
    return data.reshape(*s_size, -1, *e_size)


@verbose
def gaussian_noise_time(
    data: ndarray,
    label: ndarray,
    mean: float,
    std: float,
    seed: int = DPEEG_SEED,
    verbose=None,
) -> tuple[ndarray, ndarray]:
    """Randomly add white Gaussian noise.

    Add Gaussian noise to the raw data as new data.

    Parameters
    ----------
    data : ndarray (N, ..., T)
        Add Gaussian white noise to the last dimension of the data. Shape as
        `(N, ..., T)`, with `N` the number of data and `T` the last dimension
        of the data.
    label : ndarray (N,)
        The label corresponding to the data. Shape as `(N)`.
    mean : float
        Mean of the noise distribution.
    std : float
        Standard deviation to use for the additive noise.
    """
    size = data.shape
    data = np_flatten(data)
    rng = np.random.RandomState(seed)

    new_data = np.empty_like(data)
    for i in range(data.shape[0]):
        noise = rng.normal(loc=mean, scale=std, size=data[i].shape)
        new_data[i] = data[i] + noise
    aug_data = np.concatenate([data.reshape(size), new_data.reshape(size)])
    aug_label = np.concatenate([label, label])

    shuffle_idx = rng.permutation(aug_label.shape[0])
    aug_data = aug_data[shuffle_idx]
    aug_label = aug_label[shuffle_idx]

    return aug_data, aug_label



from .base import Sequential, SplitTrainTest, ToEEGData, split_subjects
from .trans import (
    Identity,
    Crop,
    SlideWin,
    Unsqueeze,
    Squeeze,
    Transpose,
    FilterBank,
    ApplyFunc,
    LabelMapping,
    PickLabel,
)
from .norm import (
    ZscoreNorm,
    MinMaxNorm,
)
from .aug import (
    SegRecTime,
    SlideWinAug,
    GaussTime,
)
from .metrics import (
    Pearsonr,
)
